__all__ = ('InteractionEvent',)

from scarletio import Future, copy_docs, export, future_or_timeout, include, shield

from ...bases import DiscordEntity, EventBase
from ...channel import ChannelType, create_partial_channel_from_id
from ...core import (
    APPLICATION_ID_TO_CLIENT, INTERACTION_EVENT_MESSAGE_WAITERS, INTERACTION_EVENT_RESPONSE_WAITERS, KOKORO
)
from ...message import Message
from ...permission import Permission
from ...user import ClientUserBase, ZEROUSER
from ...utils import now_as_id

from ..interaction_metadata import InteractionMetadataBase
from ..responding.constants import (
    RESPONSE_FLAG_ACKNOWLEDGED, RESPONSE_FLAG_ACKNOWLEDGING, RESPONSE_FLAG_DEFERRED, RESPONSE_FLAG_DEFERRING,
    RESPONSE_FLAG_EPHEMERAL, RESPONSE_FLAG_NONE, RESPONSE_FLAG_RESPONDED, RESPONSE_FLAG_RESPONDING
)

from .constants import (
    DEFAULT_INTERACTION_METADATA, INTERACTION_EVENT_EXPIRE_AFTER_ID_DIFFERENCE, LOCALE_DEFAULT, USER_GUILD_CACHE
)
from .fields import (
    parse_application_id, parse_application_permissions, parse_channel_id, parse_guild_id, parse_guild_locale, parse_id,
    parse_locale, parse_message, parse_token, parse_type, parse_user, parse_user_permissions, put_application_id_into,
    put_application_permissions_into, put_channel_id_into, put_guild_id_into, put_guild_locale_into, put_id_into,
    put_locale_into, put_message_into, put_token_into, put_type_into, put_user_into, put_user_permissions_into,
    validate_application_id, validate_application_permissions, validate_channel_id, validate_guild_id,
    validate_guild_locale, validate_id, validate_interaction, validate_locale, validate_message, validate_token,
    validate_type, validate_user, validate_user_permissions
)
from .preinstanced import InteractionType


create_partial_guild_from_id = include('create_partial_guild_from_id')


PRECREATE_FIELDS = {
    'application_id': ('application_id', validate_application_id),
    'application_permissions': ('application_permissions', validate_application_permissions),
    'channel_id': ('channel_id', validate_channel_id),
    'guild_id': ('guild_id', validate_guild_id),
    'guild_locale': ('guild_locale', validate_guild_locale),
    'locale': ('locale', validate_locale),
    'message': ('message', validate_message),
    'token': ('token', validate_token),
    'user': ('user', validate_user),
    'user_permissions': ('user_permissions', validate_user_permissions),
}


@export
class InteractionEvent(DiscordEntity, EventBase, immortal = True):
    """
    Represents a processed `INTERACTION_CREATE` dispatch event.
    
    Attributes
    ----------
    id : `int`
        The interaction's id.
    
    _async_task : `None`, ``Task``
        Task set if interaction event is acknowledged asynchronously.
    
    _cached_users : `None`, `list` of ``ClientUserBase``
        A list of users, which are temporary cached.
    
    _response_flag : `int`
        The response order state of ``InteractionEvent``
        
        +-------------------------------+-------+---------------------------------------------------+
        | Respective name               | Shift | Description                                       |
        +===============================+=======+===================================================+
        | RESPONSE_FLAG_DEFERRING       | 0     | The vent is being acknowledged.                   |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_DEFERRED        | 1     | The event was acknowledged and response will be   |
        |                               |       | sent later. Shows loading screen for the user.    |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_RESPONDING      | 2     | Responding to the interaction.                    |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_RESPONDED       | 3     | Response was sent on the interaction.             |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_EPHEMERAL       | 4     | Whether the main response is an ephemeral,        |
        |                               |       | showing for the invoking user only.               |
        +-------------------------------+-------+---------------------------------------------------+
        
        Can be used by extensions and is used by the the ``Client``-s to ensure correct flow order.
    
    application_id : `int`
        The interaction's application's identifier.
    
    application_permissions : ``Permission``
        The permissions granted to the application in the guild.
    
    channel_id : `int`
        The channel's identifier from where the interaction was called.
    
    guild_id : `int`
        The guild's identifier from where the interaction was called from. Might be `0` if the interaction was called
        from a private channel.
    
    guild_locale : ``Locale``
        The guild's preferred locale if invoked from guild.
    
    interaction : ``InteractionMetadataBase``
        Contain additional details of the interaction.
    
    locale : ``Locale``
        The selected language of the invoking user.
    
    message : `None`, ``Message``
        The message from where the interaction was received. Applicable for message components.
    
    token : `str`
        Interaction's token used when responding on it.
    
    type : ``InteractionType``
        The interaction's type.
    
    user : ``ClientUserBase``
        The user who called the interaction.
    
    user_permissions : ``Permission``
        The user's permissions in the respective channel.
    
    Notes
    -----
    The interaction token can be used for 15 minutes, tho if it is not used within the first 3 seconds, it is
    invalidated immediately.
    
    ˙˙InteractionEvent˙˙ instances are weakreferable.
    """
    __slots__ = (
        '_async_task', '_cached_users', '_response_flag', 'application_id', 'application_permissions', 'channel_id',
        'guild_id', 'guild_locale', 'interaction', 'locale', 'message', 'token', 'type', 'user', 'user_permissions'
    )
    
    def __new__(
        cls,
        *,
        application_id = ...,
        application_permissions = ...,
        channel_id = ...,
        guild_id = ...,
        guild_locale = ...,
        interaction = ...,
        interaction_type = ...,
        locale = ...,
        message = ...,
        token = ..., 
        user = ...,
        user_permissions = ...,
    ):
        """
        Creates a partial interaction event.
        
        Parameters
        ----------
        application_id : `int`, `str`, Optional (Keyword only)
            The interaction's application's identifier.
        
        application_permissions : ``Permission``, `int`, Optional (Keyword only)
            The permissions granted to the application in the guild.
        
        channel_id : `int`, `str`, Optional (Keyword only)
            The channel's identifier from where the interaction was called.
        
        guild_id : `int`, `str`, Optional (Keyword only)
            The guild's identifier from where the interaction was called from.
        
        guild_locale : ``Locale``, `str`, Optional (Keyword only)
            The guild's preferred locale if invoked from guild.
        
        interaction : ``InteractionMetadataBase``, Optional (Keyword only)
            Contain additional details of the interaction.
        
        locale : ``Locale``, `str`, Optional (Keyword only)
            The selected language of the invoking user.
        
        message : `None`, ``Message``, Optional (Keyword only)
            The message from where the interaction was received.
        
        token : `str`, Optional (Keyword only)
            Interaction's token used when responding on it.
        
        type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who called the interaction.
        
        user_permissions : ``Permission``, `int`, Optional (Keyword only)
            The user's permissions in the respective channel.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # application_id
        if application_id is ...:
            application_id = 0
        else:
            application_id = validate_application_id(application_id)
        
        # application_permissions
        if application_permissions is ...:
            application_permissions = Permission()
        else:
            application_permissions = validate_application_permissions(application_permissions)
        
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # guild_locale
        if guild_locale is ...:
            guild_locale = LOCALE_DEFAULT
        else:
            guild_locale = validate_guild_locale(guild_locale)
        
        # interaction & interaction_type
        if interaction_type is ...:
            interaction_type = InteractionType.none
        else:
            interaction_type = validate_type(interaction_type)
        
        if interaction is ...:
            interaction = interaction_type.metadata_type()
        else:
            interaction = validate_interaction(interaction, interaction_type)
        
        # locale
        if locale is ...:
            locale = LOCALE_DEFAULT
        else:
            locale = validate_locale(locale)
        
        # message
        if message is ...:
            message = None
        else:
            message = validate_message(message)
        
        # token
        if token is ...:
            token = ''
        else:
            token = validate_token(token)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        # user_permissions
        if user_permissions is ...:
            user_permissions = Permission()
        else:
            user_permissions = validate_user_permissions(user_permissions)
        
        # Construct
        
        self = object.__new__(cls)
        self._async_task = None
        self._cached_users = None
        self._response_flag = RESPONSE_FLAG_NONE
        self.id = 0
        self.application_id = application_id
        self.application_permissions = application_permissions
        self.type = interaction_type
        self.channel_id = channel_id
        self.guild_id = guild_id
        self.guild_locale = guild_locale
        self.interaction = interaction
        self.locale = locale
        self.token = token
        self.user = user
        self.user_permissions = user_permissions
        self.message = message
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``InteractionEvent`` with the given parameters.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            `INTERACTION_CREATE` dispatch event data.
        """
        # Need guild_id early, so we can create a guild if required.
        guild_id = parse_guild_id(data)
        
        if guild_id:
            guild = create_partial_guild_from_id(guild_id)
        else:
            guild = None
        
        # interaction_id
        interaction_id = parse_id(data)
        application_id = parse_application_id(data)
        application_permissions = parse_application_permissions(data)
        channel_id = parse_channel_id(data)
        guild_locale = parse_guild_locale(data)
        # interaction -> we will parse it later
        locale = parse_locale(data)
        message = parse_message(data)
        token = parse_token(data)
        interaction_type = parse_type(data)
        user = parse_user(data, guild_id)
        user_permissions = parse_user_permissions(data)
        
        
        self = object.__new__(cls)
        self._async_task = None
        self._cached_users = None
        self._response_flag = RESPONSE_FLAG_NONE
        self.id = interaction_id
        self.application_id = application_id
        self.application_permissions = application_permissions
        self.type = interaction_type
        self.channel_id = channel_id
        self.guild_id = guild_id
        self.guild_locale = guild_locale
        self.interaction = DEFAULT_INTERACTION_METADATA
        self.locale = locale
        self.token = token
        self.user = user
        self.user_permissions = user_permissions
        self.message = message
        
        # Cache our user if called from guild. This is required to kill references to the user in the guild.
        if (guild is not None):
            self._add_cached_user(user)
        
        # All field is set -> we can now create our own child
        self.interaction = interaction_type.metadata_type.from_data(data['data'], self)
        
        # Bind cached users to the guild for un-caching on object unallocation.
        cached_users = self._cached_users
        if (cached_users is not None):
            for user in cached_users:
                key = (user, guild)
                try:
                    reference_count = USER_GUILD_CACHE[key]
                except KeyError:
                    reference_count = 1
                else:
                    reference_count += 1
                
                USER_GUILD_CACHE[key] = reference_count
        
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Returns the interaction event's json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_application_id_into(self.application_id, data, defaults)
        put_application_permissions_into(self.application_permissions, data, defaults)
        put_channel_id_into(self.channel_id, data, defaults)
        put_guild_id_into(self.guild_id, data, defaults)
        put_guild_locale_into(self.guild_locale, data, defaults)
        put_id_into(self.id, data, defaults)
        put_locale_into(self.locale, data, defaults)
        put_message_into(self.message, data, defaults, include_internals = True)
        put_token_into(self.token, data, defaults)
        put_type_into(self.type, data, defaults)
        put_user_into(self.user, data, defaults, guild_id = self.guild_id)
        put_user_permissions_into(self.user_permissions, data, defaults)
        data['data'] = self.interaction.to_data(defaults = defaults, interaction_event = self)
        return data
    
    
    @classmethod
    def _create_empty(cls, interaction_id):
        """
        Creates a new partial interaction event with it's attributes set as their default value.
        
        Parameters
        ----------
        interaction_id : `int`
            The interaction event's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._async_task = None
        self._cached_users = None
        self._response_flag = RESPONSE_FLAG_NONE
        self.id = interaction_id
        self.application_id = 0
        self.application_permissions = Permission()
        self.type = InteractionType.none
        self.channel_id = 0
        self.guild_id = 0
        self.guild_locale = LOCALE_DEFAULT
        self.interaction = DEFAULT_INTERACTION_METADATA
        self.locale = LOCALE_DEFAULT
        self.token = ''
        self.user = ZEROUSER
        self.user_permissions = Permission()
        self.message = None
        return self
    
    
    @classmethod
    def precreate(cls, interaction_id, **keyword_parameters):
        """
        Creates an interaction event. Not like ``.__new__``, ``.precreate`` allows setting ``.id`` as well.
        
        Since interaction events are not cached, there os no major advantage of using this method.
        
        Parameters
        ----------
        interaction_id : `int`
            The interaction's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters defining which attribute and how should be set.
        
        Other Parameters
        ----------
        application_id : `int`, `str`, Optional (Keyword only)
            The interaction's application's identifier.
        
        application_permissions : ``Permission``, `int`, Optional (Keyword only)
            The permissions granted to the application in the guild.
        
        channel_id : `int`, `str`, Optional (Keyword only)
            The channel's identifier from where the interaction was called.
        
        guild_id : `int`, `str`, Optional (Keyword only)
            The guild's identifier from where the interaction was called from.
        
        guild_locale : ``Locale``, `str`, Optional (Keyword only)
            The guild's preferred locale if invoked from guild.
        
        interaction : ``InteractionMetadataBase``, Optional (Keyword only)
            Contain additional details of the interaction.
        
        locale : ``Locale``, `str`, Optional (Keyword only)
            The selected language of the invoking user.
        
        message : `None`, ``Message``, Optional (Keyword only)
            The message from where the interaction was received.
        
        token : `str`, Optional (Keyword only)
            Interaction's token used when responding on it.
        
        type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who called the interaction.
        
        user_permissions : ``Permission``, `int`, Optional (Keyword only)
            The user's permissions in the respective channel.
            
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        interaction_id = validate_id(interaction_id)
        
        if keyword_parameters:
            processable = []
            
            # interaction & interaction_type
            try:
                interaction_type = keyword_parameters.pop('interaction_type')
            except KeyError:
                interaction_type = InteractionType.none
            else:
                interaction_type = validate_type(interaction_type)
                processable.append(('type', interaction_type))
            
            try:
                interaction = keyword_parameters.pop('interaction')
            except KeyError:
                if interaction_type is not InteractionType.none:
                    interaction = interaction_type.metadata_type()
                    processable.append(('interaction', interaction))
            else:
                interaction = validate_interaction(interaction, interaction_type)
                processable.append(('interaction', interaction))
            
            
            extra = None
            
            while keyword_parameters:
                field_name, field_value = keyword_parameters.popitem() 
                try:
                    attribute_name, validator = PRECREATE_FIELDS[field_name]
                except KeyError:
                    if extra is None:
                        extra = {}
                    extra[field_name] = field_value
                    continue
                
                attribute_value = validator(field_value)
                processable.append((attribute_name, attribute_value))
                continue
                
            if (extra is not None):
                raise TypeError(
                    f'Unused or unsettable keyword parameters: {extra!r}.'
                )
        
        else:
            processable = None
        
        self = cls._create_empty(interaction_id)
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self
    
    
    def copy(self):
        """
        Copies the interaction event.
        The returned interaction event is partial lacking internal fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new._async_task = None
        new._cached_users = None
        new._response_flag = RESPONSE_FLAG_NONE
        new.id = 0
        new.application_id = self.application_id
        new.application_permissions = self.application_permissions
        new.type = self.type
        new.channel_id = self.channel_id
        new.guild_id = self.guild_id
        new.guild_locale = self.guild_locale
        new.interaction = self.interaction.copy()
        new.locale = self.locale
        new.token = self.token
        new.user = self.user
        new.user_permissions = self.user_permissions
        new.message = self.message
        return new
    
    
    def copy_with(
        self,
        *,
        application_id = ...,
        application_permissions = ...,
        channel_id = ...,
        guild_id = ...,
        guild_locale = ...,
        interaction = ...,
        interaction_type = ...,
        locale = ...,
        message = ...,
        token = ..., 
        user = ...,
        user_permissions = ...,
    ):
        """
        Copies the interaction event modifying it's defined fields.
        The returned interaction event is partial lacking internal fields.
        
        Parameters
        ----------
        application_id : `int`, `str`, Optional (Keyword only)
            The interaction's application's identifier.
        
        application_permissions : ``Permission``, `int`, Optional (Keyword only)
            The permissions granted to the application in the guild.
        
        channel_id : `int`, `str`, Optional (Keyword only)
            The channel's identifier from where the interaction was called.
        
        guild_id : `int`, `str`, Optional (Keyword only)
            The guild's identifier from where the interaction was called from.
        
        guild_locale : ``Locale``, `str`, Optional (Keyword only)
            The guild's preferred locale if invoked from guild.
        
        interaction : ``InteractionMetadataBase``, Optional (Keyword only)
            Contain additional details of the interaction.
        
        locale : ``Locale``, `str`, Optional (Keyword only)
            The selected language of the invoking user.
        
        message : `None`, ``Message``, Optional (Keyword only)
            The message from where the interaction was received.
        
        token : `str`, Optional (Keyword only)
            Interaction's token used when responding on it.
        
        type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who called the interaction.
        
        user_permissions : ``Permission``, `int`, Optional (Keyword only)
            The user's permissions in the respective channel.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        # application_id
        if application_id is ...:
            application_id = self.application_id
        else:
            application_id = validate_application_id(application_id)
        
        # application_permissions
        if application_permissions is ...:
            application_permissions = self.application_permissions
        else:
            application_permissions = validate_application_permissions(application_permissions)
        
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # guild_locale
        if guild_locale is ...:
            guild_locale = self.guild_locale
        else:
            guild_locale = validate_guild_locale(guild_locale)
        
        # interaction & interaction_type
        if interaction_type is ...:
            interaction_type = self.type
        else:
            interaction_type = validate_type(interaction_type)
        
        if interaction is ...:
            if interaction_type is self.type:
                interaction = self.interaction.copy()
            else:
                interaction = interaction_type.metadata_type()
        else:
            interaction = validate_interaction(interaction, interaction_type)
        
        # locale
        if locale is ...:
            locale = self.locale
        else:
            locale = validate_locale(locale)
        
        # message
        if message is ...:
            message = self.message
        else:
            message = validate_message(message)
        
        # token
        if token is ...:
            token = self.token
        else:
            token = validate_token(token)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # user_permissions
        if user_permissions is ...:
            user_permissions = self.user_permissions
        else:
            user_permissions = validate_user_permissions(user_permissions)
        
        new = object.__new__(type(self))
        new._async_task = None
        new._cached_users = None
        new._response_flag = RESPONSE_FLAG_NONE
        new.id = 0
        new.application_id = application_id
        new.application_permissions = application_permissions
        new.type = interaction_type
        new.channel_id = channel_id
        new.guild_id = guild_id
        new.guild_locale = guild_locale
        new.interaction = interaction
        new.locale = locale
        new.token = token
        new.user = user
        new.user_permissions = user_permissions
        new.message = message
        return new
    
    
    async def wait_for_response_message(self, *, timeout = None):
        """
        Waits for response message. Applicable for application command interactions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        timeout : `None`, `float` = `None`, Optional (Keyword only)
            The maximal time to wait for message before `TimeoutError` is raised.
        
        Returns
        -------
        message : ``Message``
            The received message.
        
        Raises
        ------
        RuntimeError
            The interaction was acknowledged with `show_for_invoking_user_only=True` (as ephemeral). Response
            `message` cannot be detected.
        TimeoutError
            Message was not received before timeout.
        """
        message = self.message
        if (message is not None):
            return message
        
        if self._response_flag & RESPONSE_FLAG_EPHEMERAL:
            raise RuntimeError(
                f'The interaction was acknowledged with `show_for_invoking_user_only = True` '
                f'(as ephemeral). Response `message` cannot be detected.'
            )
        
        try:
            waiter = INTERACTION_EVENT_MESSAGE_WAITERS[self]
        except KeyError:
            waiter = Future(KOKORO)
            INTERACTION_EVENT_MESSAGE_WAITERS[self] = waiter
        
        waiter = shield(waiter, KOKORO)
        
        if (timeout is not None):
            future_or_timeout(waiter, timeout)
        
        await waiter
        return self.message
    
    
    def __del__(self):
        """
        Unregisters the user-guild pair from the interaction cache.
        """
        cached_users = self._cached_users
        if cached_users is None:
            return
        
        guild = self.guild
        if (guild is None):
            return
        
        for user in cached_users:
            key = (user, guild)
            
            # A client meanwhile joined the guild?
            if not guild.partial:
                try:
                    del USER_GUILD_CACHE[key]
                except KeyError:
                    pass
                return
            
            try:
                reference_count = USER_GUILD_CACHE[key]
            except KeyError:
                reference_count = 0
            else:
                if reference_count == 1:
                    del USER_GUILD_CACHE[key]
                    reference_count = 0
                else:
                    reference_count -= 1
            
            if reference_count == 0:
                try:
                    del user.guild_profiles[guild.id]
                except KeyError:
                    pass
    
    
    def __repr__(self):
        """Returns the representation of the event."""
        repr_parts = ['<', self.__class__.__name__]
        
        response_state_names = None
        response_state = self._response_flag
        
        if response_state == RESPONSE_FLAG_NONE:
            pass
        
        if response_state & RESPONSE_FLAG_DEFERRING:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('deferring')
        
        if response_state & RESPONSE_FLAG_DEFERRED:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('deferred')
        
        if response_state & RESPONSE_FLAG_RESPONDING:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('responding')
        
        if response_state & RESPONSE_FLAG_RESPONDED:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('responded')
        
        if response_state & RESPONSE_FLAG_EPHEMERAL:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('ephemeral')
        
        if (response_state_names is not None):
            repr_parts.append(' (')
            index = 0
            limit = len(response_state_names)
            while True:
                response_state_name = response_state_names[index]
                repr_parts.append(response_state_name)
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append('),')
        
        repr_parts.append(' type = ')
        metadata_type = self.type
        repr_parts.append(metadata_type.name)
        repr_parts.append(' (')
        repr_parts.append(repr(metadata_type.value))
        repr_parts.append(')')
        
        
        guild_id = self.guild_id
        if guild_id:
            repr_parts.append(', guild_id = ')
            repr_parts.append(repr(guild_id))
            
            repr_parts.append(', application_permissions = ')
            repr_parts.append(format(self.application_permissions, 'd'))
        
        
        repr_parts.append(', channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        
        message = self.message
        if (message is not None):
            repr_parts.append(', message = ')
            repr_parts.append(repr(message))
        
        
        repr_parts.append(', user = ')
        repr_parts.append(repr(self.user))
        
        repr_parts.append(', guild_locale = ')
        repr_parts.append(repr(self.guild_locale.name))
        
        if guild_id:
            repr_parts.append(', locale = ')
            repr_parts.append(repr(self.locale.name))
        
        repr_parts.append(', interaction = ')
        repr_parts.append(repr(self.interaction))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two interaction events are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # Shortcut
        if self is other:
            return True
        
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id and (self.id != other.id):
            return False
        
        # application_id
        if self.application_id != other.application_id:
            return False
        
        # application_permissions
        if self.application_permissions != other.application_permissions:
            return False
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # guild_locale
        if self.guild_locale is not other.guild_locale:
            return False
        
        # interaction
        if self.interaction != other.interaction:
            return False
        
        # locale
        if self.locale is not other.locale:
            return False
        
        # message
        if self.message is not other.message:
            return False
        
        # token
        if self.token != other.token:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # user
        if self.user is not other.user:
            return False
        
        # user_permissions
        if self.user_permissions != other.user_permissions:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the interaction event."""
        hash_value = self.id
        if hash_value and (self._cached_users is not None):
            return hash_value
        
        # application_id
        hash_value ^= self.application_id
        
        # application_permissions
        hash_value ^= self.application_permissions
        
        # channel_id
        hash_value ^= self.channel_id
        
        # guild_id
        hash_value ^= self.guild_id
        
        # guild_locale
        hash_value ^= hash(self.guild_locale)
        
        # interaction
        hash_value ^= hash(self.interaction)
        
        # locale
        hash_value ^= hash(self.locale)
        
        # message
        message = self.message
        if (message is not None):
            hash_value ^= hash(self.message)
        
        # token
        hash_value ^= hash(self.token)
        
        # type
        hash_value ^= self.type.value
        
        # user
        hash_value ^= hash(self.user)
        
        # user_permissions
        hash_value ^= self.user_permissions
        
        return hash_value
    
    
    def is_unanswered(self):
        """
        Returns whether the event was not acknowledged and not acknowledging either.
        
        Returns
        -------
        is_unanswered : `bool`
        """
        return True if (self._response_flag == RESPONSE_FLAG_NONE) else False
    
    
    def is_acknowledging(self):
        """
        Returns whether the event is being acknowledged.
        
        Returns
        -------
        is_acknowledging : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_ACKNOWLEDGING) else False
    
    
    def is_acknowledged(self):
        """
        Returns whether the event is acknowledged.
        
        Returns
        -------
        is_acknowledged : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_ACKNOWLEDGED) else False
    
    
    def is_deferred(self):
        """
        Returns whether the event is deferred.
        
        Returns
        -------
        is_deferred : `bool`
        """
        response_state = self._response_flag
        if response_state & RESPONSE_FLAG_RESPONDED:
            return False
        
        if response_state & RESPONSE_FLAG_DEFERRED:
            return True
        
        return False
    
    
    def is_responding(self):
        """
        Returns whether the event it being responded.
        
        Returns
        -------
        is_responding : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_RESPONDING) else False
    
    
    def is_responded(self):
        """
        Returns whether was responded.
        
        Returns
        -------
        is_responded : `bool`
        """
        return True if (self._response_flag & RESPONSE_FLAG_RESPONDED) else False
    
    
    def is_expired(self):
        """
        Returns whether the event is already expired.
        
        Returns
        -------
        is_expired : `bool`
        """
        return now_as_id() > (self.id + INTERACTION_EVENT_EXPIRE_AFTER_ID_DIFFERENCE)
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 3
    
    
    def __iter__(self):
        """
        Unpacks the event.
        
        This method is a generator.
        """
        yield self.type
        yield self.user
        yield self.interaction
    
    
    @property
    def channel(self):
        """
        Returns the interaction's event's channel.
        
        Returns
        -------
        channel : ``Channel``, `None`
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the interaction's guild.
        
        Returns
        -------
        guild : ``Guild``, `None`
        """
        guild_id = self.guild_id
        if guild_id:
            return create_partial_guild_from_id(guild_id)
    
    
    @property
    def client(self):
        """
        Returns the interaction's client.
        
        Returns
        -------
        client : `Client`
        
        Raises
        ------
        RuntimeError
            Client could not be identified.
        """
        try:
            return APPLICATION_ID_TO_CLIENT[self.application_id]
        except KeyError:
            raise RuntimeError(
                f'Client of {self!r} could not be identified.'
            ) from None
    
    
    @property
    def voice_client(self):
        """
        Returns the voice client of the interaction's client in it's guild.
        
        Returns
        -------
        voice_client : `None`, ``VoiceClient``
        """
        try:
            client = APPLICATION_ID_TO_CLIENT[self.application_id]
        except KeyError:
            voice_client = None
        else:
            guild_id = self.message.guild_id
            if guild_id:
                voice_client = client.voice_clients.get(guild_id, None)
            else:
                voice_client = None
        
        return voice_client
    
    
    def _add_cached_user(self, user):
        """
        Adds a user to the cached ones by the interaction.
        
        This function might be called inside of ``InteractionEvent.__new__`` when initializing it's interaction.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user to add to cache.
        """
        cached_users = self._cached_users
        if cached_users is None:
            self._cached_users = [user]
        else:
            if (user not in cached_users):
                cached_users.append(user)
    
    
    def _add_response_waiter(self):
        """
        Adds the interaction event to response waiters.
        
        Called when the interaction is application command one, to resolve it's ``.message`` attribute when created.
        """
        INTERACTION_EVENT_RESPONSE_WAITERS[self.id] = self
    
    
    async def _wait_for_async_task_completion(self):
        """
        Waits for async task's completion.
        
        This method is a coroutine.
        """
        async_task = self._async_task
        if (async_task is not None) and (async_task is not KOKORO.current_task):
            try:
                await async_task
            finally:
                self._async_task = None
    
    # Field Proxies
    
    @property
    @copy_docs(InteractionMetadataBase.component_type)
    def component_type(self):
        return self.interaction.component_type
    
    
    @property
    @copy_docs(InteractionMetadataBase.components)
    def components(self):
        return self.interaction.components
    
    
    @property
    @copy_docs(InteractionMetadataBase.custom_id)
    def custom_id(self):
        return self.interaction.custom_id
    
    
    @property
    @copy_docs(InteractionMetadataBase.id)
    def application_command_id(self):
        return self.interaction.id
    
    
    @property
    @copy_docs(InteractionMetadataBase.name)
    def application_command_name(self):
        return self.interaction.name
    
    
    @property
    @copy_docs(InteractionMetadataBase.options)
    def options(self):
        return self.interaction.options
    
    
    @property
    @copy_docs(InteractionMetadataBase.resolved)
    def resolved(self):
        return self.interaction.resolved
    
    
    @property
    @copy_docs(InteractionMetadataBase.target_id)
    def target_id(self):
        return self.interaction.target_id
    
    
    @property
    @copy_docs(InteractionMetadataBase.values)
    def values(self):
        return self.interaction.values
    
    # application command
    
    @property
    @copy_docs(InteractionMetadataBase.target)
    def target(self):
        return self.interaction.target
    
    # application command autocomplete
    
    @copy_docs(InteractionMetadataBase.iter_options)
    def iter_options(self):
        return (yield from self.interaction.iter_options())
    
    
    @property
    @copy_docs(InteractionMetadataBase.focused_option)
    def focused_option(self):
        return self.interaction.focused_option
    
    
    @copy_docs(InteractionMetadataBase.get_non_focused_values)
    def get_non_focused_values(self):
        return self.interaction.get_non_focused_values()


    @copy_docs(InteractionMetadataBase.get_value_of)
    def get_value_of(self, *option_names):
        return self.interaction.get_value_of(*option_names)
    
    
    @property
    @copy_docs(InteractionMetadataBase.value)
    def value(self):
        return self.interaction.value
    
    # Message component
    
    @copy_docs(InteractionMetadataBase.iter_values)
    def iter_values(self):
        return (yield from self.interaction.iter_values())
    
    
    @copy_docs(InteractionMetadataBase.iter_entities)
    def iter_entities(self):
        return (yield from self.interaction.iter_entities())
    
    
    @property
    @copy_docs(InteractionMetadataBase.entities)
    def entities(self):
        return self.interaction.entities
    
    # form submit
    
    @copy_docs(InteractionMetadataBase.iter_components)
    def iter_components(self):
        return (yield from self.interaction.iter_components())
    
    
    @copy_docs(InteractionMetadataBase.iter_custom_ids_and_values)
    def iter_custom_ids_and_values(self):
        return (yield from self.interaction.iter_custom_ids_and_values())
    
    
    @copy_docs(InteractionMetadataBase.get_custom_id_value_relation)
    def get_custom_id_value_relation(self):
        return self.interaction.get_custom_id_value_relation()
    
    
    @copy_docs(InteractionMetadataBase.get_value_for)
    def get_value_for(self, custom_id_to_match):
        return self.interaction.get_value_for(custom_id_to_match)
    
    
    @copy_docs(InteractionMetadataBase.get_match_and_value)
    def get_match_and_value(self, matcher):
        return self.interaction.get_match_and_value(matcher)
    
    
    @copy_docs(InteractionMetadataBase.iter_matches_and_values)
    def iter_matches_and_values(self, matcher):
        return (yield from self.interaction.iter_matches_and_values(matcher))
    
    # resolved
    
    @copy_docs(InteractionMetadataBase.resolve_attachment)
    def resolve_attachment(self, attachment_id):
        return self.interaction.resolve_attachment(attachment_id)
    
    
    @copy_docs(InteractionMetadataBase.resolve_channel)
    def resolve_channel(self, channel_id):
        return self.interaction.resolve_channel(channel_id)
    
    
    @copy_docs(InteractionMetadataBase.resolve_message)
    def resolve_message(self, message_id):
        return self.interaction.resolve_message(message_id)
    
    
    @copy_docs(InteractionMetadataBase.resolve_role)
    def resolve_role(self, role_id):
        return self.interaction.resolve_role(role_id)
    
    
    @copy_docs(InteractionMetadataBase.resolve_user)
    def resolve_user(self, user_id):
        return self.interaction.resolve_user(user_id)
    
    
    @copy_docs(InteractionMetadataBase.resolve_mentionable)
    def resolve_mentionable(self, mentionable_id):
        return self.interaction.resolve_mentionable(mentionable_id)
    
    
    @copy_docs(InteractionMetadataBase.resolve_entity)
    def resolve_entity(self, entity_id):
        return self.interaction.resolve_entity(entity_id)



__all__ = ('AutoModerationRuleTriggerMetadataKeyword',)

from scarletio import copy_docs

from .base import AutoModerationRuleTriggerMetadataBase
from .fields import (
    parse_excluded_keywords, parse_keywords, parse_regex_patterns, put_excluded_keywords_into, put_keywords_into,
    put_regex_patterns_into, validate_excluded_keywords, validate_keywords, validate_regex_patterns
)


class AutoModerationRuleTriggerMetadataKeyword(AutoModerationRuleTriggerMetadataBase):
    """
    Keyword trigger metadata for an auto moderation rule.
    
    Attributes
    ----------
    excluded_keywords : `None`, `tuple` of `str`
        Excluded keywords from under the rule.
    
    keywords : `None`, `tuple` of `str`
        Substrings which will be searched for in content.
    
    regex_patterns : `None`, `tuple` of `str`
        Regular expression patterns which are matched against content.
        
        > Only rust flavored regex is supported.
    
    Matching Strategies
    -------------------
    Use the `*` wildcard symbol at the beginning and end of a keyword to define how the keyword should be matched.
    All keywords are not case sensitive.
    
    - `Prefix` - word must start with the keyword
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | cat*      | catch, Catapult, CAttLE               |
        +-----------+---------------------------------------+
        | tra*      | train, trade, TRAditional             |
        +-----------+---------------------------------------+
        | the mat*  | the matrix                            |
        +-----------+---------------------------------------+
    
    
    - `Suffix` - word must end with the keyword
    
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | *cat      | wildcat, copyCat                      |
        +-----------+---------------------------------------+
        | *tra      | extra, ultra, orchesTRA               |
        +-----------+---------------------------------------+
        | *the mat  | breathe mat                           |
        +-----------+---------------------------------------+
    
    
    - `Anywhere` - keyword can appear anywhere in the content
    
        +-----------+---------------------------------------+
        | Keyword     | Matches                             |
        +===========+=======================================+
        | *cat*       | location, eduCation                 |
        +-----------+---------------------------------------+
        | *tra*       | abstracted, outrage                 |
        +-----------+---------------------------------------+
        | *the mat*   | breathe matter                      |
        +-----------+---------------------------------------+
    
    
    - `Whole Word` - keyword is a full word or phrase and must be surrounded by whitespace at the beginning and end
    
        +-----------+---------------------------------------+
        | Keyword   | Matches                               |
        +===========+=======================================+
        | cat       | cat                                   |
        +-----------+---------------------------------------+
        | train     | train                                 |
        +-----------+---------------------------------------+
        | the mat   | the mat                               |
        +-----------+---------------------------------------+
    """
    __slots__ = ('excluded_keywords', 'keywords', 'regex_patterns')
    
    def __new__(cls, keywords = None, regex_patterns = None, excluded_keywords = None):
        """
        Creates a new keyword trigger metadata for ``AutoModerationRule``-s.
        
        Parameters
        ----------
        keywords : `None`, `str`, `iterable` of `str` = `None`, Optional
            Substrings which will be searched for in content.
        
        regex_patterns : `None`, `tuple` of `str` = `None`, Optional
            Regular expression patterns which are matched against content.
        
        excluded_keywords : `None`, `str`, `iterable` of `str` = `None`, Optional
            Excluded keywords from the preset.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        excluded_keywords = validate_excluded_keywords(excluded_keywords)
        keywords = validate_keywords(keywords)
        regex_patterns = validate_regex_patterns(regex_patterns)
        
        self = object.__new__(cls)
        self.excluded_keywords = excluded_keywords
        self.keywords = keywords
        self.regex_patterns = regex_patterns
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # keywords
        keywords = self.keywords
        repr_parts.append(' keywords = [')
        if (keywords is not None):
            index = 0
            limit = len(keywords)
            while True:
                keyword = keywords[index]
                repr_parts.append(repr(keyword))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        # regex_patterns
        regex_patterns = self.regex_patterns
        repr_parts.append(', regex_patterns = [')
        if (regex_patterns is not None):
            index = 0
            limit = len(regex_patterns)
            while True:
                keyword = regex_patterns[index]
                repr_parts.append(repr(keyword))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        repr_parts.append(', excluded_keywords = [')
        if (excluded_keywords is not None):
            index = 0
            limit = len(excluded_keywords)
            while True:
                excluded_keyword = excluded_keywords[index]
                repr_parts.append(repr(excluded_keyword))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationRuleTriggerMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.excluded_keywords = parse_excluded_keywords(data)
        self.keywords = parse_keywords(data)
        self.regex_patterns = parse_regex_patterns(data)
        return self
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.to_data)
    def to_data(self, defaults = False):
        data = {}
        put_excluded_keywords_into(self.excluded_keywords, data, defaults)
        put_keywords_into(self.keywords, data, defaults)
        put_regex_patterns_into(self.regex_patterns, data, defaults)
        return data
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # excluded_keywords
        if self.excluded_keywords != other.excluded_keywords:
            return False
        
        # keywords
        if self.keywords != other.keywords:
            return False
        
        # regex_patterns
        if self.regex_patterns != other.regex_patterns:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        if (excluded_keywords is not None):
            hash_value ^= len(excluded_keywords) << 8
            
            for excluded_keyword in excluded_keywords:
                hash_value ^= hash(excluded_keyword)
        
        # keywords
        keywords = self.keywords
        if (keywords is not None):
            hash_value ^= len(keywords)
            
            for keyword in keywords:
                hash_value ^= hash(keyword)
        
        # regex_patterns
        regex_patterns = self.regex_patterns
        if (regex_patterns is not None):
            hash_value ^= len(regex_patterns) << 4
            
            for regex_pattern in regex_patterns:
                hash_value ^= hash(regex_pattern)
        
        return hash_value
    
    
    @copy_docs(AutoModerationRuleTriggerMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # excluded_keywords
        excluded_keywords = self.excluded_keywords
        if (excluded_keywords is not None):
            excluded_keywords = (*excluded_keywords, )
        new.excluded_keywords = excluded_keywords
        
        # keywords
        keywords = self.keywords
        if (keywords is not None):
            keywords = (*keywords, )
        new.keywords = keywords
        
        # regex_patterns
        regex_patterns = self.regex_patterns
        if (regex_patterns is not None):
            regex_patterns = (*regex_patterns, )
        new.regex_patterns = regex_patterns
        
        return new

    
    def copy_with(self, *, excluded_keywords = ..., keywords = ..., regex_patterns = ...):
        """
        Copies the trigger metadata with altering it's attributes based on the given fields.
        
        Parameters
        ----------
        excluded_keywords : `None`, `str`, `iterable` of `str, Optional (Keyword only)
            Excluded keywords from the rule.
        
        keywords : `None`, `str`, `iterable` of `str`, Optional (Keyword only)
            Substrings which will be searched for in content.
        
        regex_patterns : `None`, `tuple` of `str`, Optional (Keyword only)
            Regular expression patterns which are matched against content.
        
        Returns
        -------
        new : `instance<type<self>>`
               
        Raises
        ------
        TypeError
            - If a parameter of incorrect type is given.
        """
        # excluded_keywords
        if excluded_keywords is ...:
            excluded_keywords = self.excluded_keywords
            if (excluded_keywords is not None):
                excluded_keywords = (*excluded_keywords, )
        else:
            excluded_keywords = validate_excluded_keywords(excluded_keywords)
        
        # keywords
        if keywords is ...:
            keywords = self.keywords
            if (keywords is not None):
                keywords = (*keywords, )
        else:
            keywords = validate_keywords(keywords)
        
        # regex_patterns
        if regex_patterns is ...:
            regex_patterns = self.regex_patterns
            if (regex_patterns is not None):
                regex_patterns = (*regex_patterns, )
        else:
            regex_patterns = validate_regex_patterns(regex_patterns)
        
        new = object.__new__(type(self))
        new.excluded_keywords = excluded_keywords
        new.keywords = keywords
        new.regex_patterns = regex_patterns
        return new



__all__ = ()

import reprlib
from collections import deque
from datetime import datetime, timedelta
from math import floor
from os.path import split as split_path

from scarletio import to_json
from scarletio.web_common import Formdata

from ...env import API_VERSION

from ..application_command import ApplicationCommand
from ..auto_moderation import AutoModerationRule
from ..bases import maybe_snowflake, maybe_snowflake_pair, maybe_snowflake_token_pair
from ..channel import Channel, ForumTag, PermissionOverwrite
from ..component import Component, ComponentType, create_row
from ..core import (
    APPLICATION_COMMANDS, AUTO_MODERATION_RULES, CHANNELS, FORUM_TAGS, GUILDS, MESSAGES, ROLES, SCHEDULED_EVENTS,
    STICKERS, STICKER_PACKS, USERS
)
from ..embed import EmbedBase
from ..emoji import Emoji, parse_reaction
from ..guild import Guild
from ..message import Attachment, Message
from ..oauth2 import Achievement, Oauth2Access, Oauth2User
from ..role import Role
from ..scheduled_event import ScheduledEvent
from ..stage import Stage
from ..sticker import Sticker, StickerPack
from ..user import ClientUserBase
from ..utils import random_id
from ..webhook import Webhook


def get_components_data(components, is_edit):
    """
    Gets component data from the given components.
    
    Parameters
    ----------
    components : `None`, ``Component``, (`set`, `list`) of (``Component``, (`set`, `list`) of ``Component``)
        Components to be attached to a message.
    is_edit : `bool`
        Whether the processed `components` fields are for message edition. At this case passing `None` will
        remove them.
    
    Returns
    -------
    component_datas : `None`, `list` of (`dict` of (`str`, `Any`) items)
        The generated data if any.
    
    Raises
    ------
    TypeError
        - If `components` was not given neither as `None`, ``Component``, (`list`, `tuple`) of ``Component``
            instances.
    AssertionError
        - If `components` contains a non ``Component`` element.
    """
    
    # Components check order:
    # 1.: None -> None || []
    # 2.: Ellipsis -> None || Ellipsis
    # 2.: Component -> [component.to_data()]
    # 3.: (list, tuple) of Component, (list, tuple) of Component -> [component.to_data(), ...] / None
    # 4.: raise
    
    if components is None:
        if is_edit:
            component_datas = []
        else:
            component_datas = None
    
    elif components is ...:
        if is_edit:
            component_datas = ...
        else:
            component_datas = None
    
    else:
        if isinstance(components, Component):
            if components.type is not ComponentType.row:
                components = create_row(components)
                
            component_datas = [components.to_data()]
        elif isinstance(components, (list, tuple)):
            component_datas = None
            
            for component in components:
                if isinstance(component, Component):
                    if component.type is not ComponentType.row:
                        component = create_row(component)
                
                elif isinstance(component, (list, tuple)):
                    component = create_row(*component)
                
                else:
                    raise TypeError(
                        f'`components` can contain contain `{Component.__name__}`, (`list`, `tuple`) of '
                        f'`{Component.__name__}`, got {components.__class__.__name__}; {components!r}.'
                    )
                
                if component_datas is None:
                    component_datas = []
                
                component_datas.append(component.to_data())
                continue
        
        else:
            raise TypeError(
                f'`components` can be `{Component.__name__}`, (`list`, `tuple`) of '
                f'(`{Component.__name__}`, (`list`, `tuple`) of `{Component.__name__}`), '
                f'got {components.__class__.__name__}; {components!r}.'
            )
    
    return component_datas


def validate_message_to_delete(message):
    """
    Validates a message to delete.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message to validate for deletion.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier where the message is.
    message_id : `int`
        The message's identifier.
    message : `None`, ``Message``
        The referenced message if found.
    
    Raises
    ------
    TypeError
        If message was not given neither as ``Message``, `tuple` (`int`, `int`).
    """
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {message.__class__.__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
        
        message = MESSAGES.get(message, None)
    
    return channel_id, message_id, message


if API_VERSION >= 9:
    def _add_file_from_tuple(files, tuple_):
        """
        Processes the given file tuple.
        
        Parameters
        ----------
        files : `None`, `list` of `tuple` (`str`, `Any`, (`None`, `str`))
            The collected files to send.
        tuple_ : `tuple` (?(`None`, `str`), ?`Any`, ?(`None`, `str`))
            A tuple containing the name, io and the description.
        
        Returns
        -------
        files : `None`, `list` of `tuple` (`str`, `Any`, (`None`, `str`))
            The collected files to send.
        
        Raises
        ------
        ValueError
            If a `tuple_`'s length is out of the expected [0:3] range.
        """
        tuple_length = len(tuple_)
        if tuple_length:
            if tuple_length > 3:
                raise ValueError(
                    f'`tuple` length can be in range [0:3], got {tuple_length!r}; {tuple_!r}.'
                )
            
            if tuple_length == 1:
                io = tuple_[0]
                name = None
                description = None
            elif tuple_length == 2:
                name, io = tuple_
                if (name is not None) and (not name):
                    name = None
                description = None
            else:
                name, io, description = tuple_
                if (name is not None) and (not name):
                    name = None
                if (description is not None) and (not description):
                    description = None
            
            if name is None:
                name = getattr(io, 'name', None)
                if (name is not None) and name:
                    _, name = split_path(name)
                else:
                    name = str(random_id())
            
            if files is None:
                files = []
            
            files.append((name, io, description))
        
        return files
    
    
    def _build_partial_attachment_data(attachment_id, description):
        """
        Builds a partial attachment payload to be sent to Discord.
        
        Parameters
        ----------
        attachment_id : `int`
            The attachment's identifier.
        description : `None`, `str`
            Description for the attachment.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`)
        """
        data = {'id': attachment_id}
        
        if (description is not None):
            data['description'] = description
        
        return data
    
    
    def create_file_form(data, file):
        """
        Creates a `multipart/form-data` form from the message's data and from the file data. If there is no files to
        send, will return `None` to tell the caller, that nothing is added to the overall data.
        
        Parameters
        ----------
        data : `dict` of `Any`
            The data created by the ``.message_create`` method.
        file : `dict` of (`file-name`, `io`) items, `list` of (``Attachment``, `tuple` (`file-name`, `io`)) elements,
                `tuple` (`file-name`, `io`, ), `io`, ``Attachment``
            The files to send.
        
        Returns
        -------
        form : `None`, `Formdata`
            Returns a `Formdata` of the files and from the message's data. If there are no files to send, returns
            `None` instead.
        contains_attachments : `bool`
            Whether the message payload contains attachments.
        
        Raises
        ------
        ValueError
            - If more than `10` file is registered to be sent.
            - If a `tuple`'s length is out of the expected [0:3] range.
        
        Notes
        -----
        Accepted `io` types with check order are:
        - ``BodyPartReader``
        - `bytes`, `bytearray`, `memoryview`
        - `str`
        - `BytesIO`
        - `StringIO`
        - `TextIOBase`
        - `BufferedReader`, `BufferedRandom`
        - `IOBase`
        - ``AsyncIO``
        - `async-iterable`
        
        Raises `TypeError` at the case of invalid `io` type.
        
        There are two predefined data types specialized to send files:
        - ``ReuBytesIO``
        - ``ReuAsyncIO``
        
        If a buffer is sent, then when the request is done, it is closed. So if the request fails, we would not be
        able to resend the file, except if we have a data type, what instead of closing on `.close()` just seeks to
        `0` (or later if needed) on close, instead of really closing instantly. These data types implement a
        `.real_close()` method, but they do `real_close` on `__exit__` as well.
        """
        files = None
        attachments = None
        
        # checking structure
        
        # case 0 None
        if file is None:
            pass
        
        # case 1 dict like
        elif hasattr(type(file), 'items'):
            for name, io in file.items():
                if files is None:
                    files = []
                
                files.append((name, io, None))
        
        # case 2 tuple => file, filename pair | file, filename, description pairs
        elif isinstance(file, tuple):
            files = _add_file_from_tuple(files, file)
        
        # case 3 list like
        elif isinstance(file, (list, deque)):
            for element in file:
                if isinstance(element, Attachment):
                    if attachments is None:
                        attachments = []
                    
                    attachments.append(element.id)
                
                else:
                    files = _add_file_from_tuple(files, element)
        
        elif isinstance(file, Attachment):
            if attachments is None:
                attachments = []
            
            attachments.append(file.id)
        
        # case 4 file itself
        else:
            name = getattr(file, 'name', None)
            if (name is not None) and name:
                _, name = split_path(name)
            else:
                name = str(random_id())
            
            if files is None:
                files = []
            
            files.append((name, file, None),)
        
        attachment_fields = None
        
        if (files is not None):
            for index, (name, io, description) in enumerate(files):
                if attachment_fields is None:
                    attachment_fields = []
                
                attachment_fields.append(_build_partial_attachment_data(index, description))
        
        if (attachments is not None):
            for attachment_id in attachments:
                if attachment_fields is None:
                    attachment_fields = []
                
                attachment_fields.append(_build_partial_attachment_data(attachment_id, None))
        
        if (attachment_fields is None):
            contains_attachments = False
            
            form = None
        else:
            contains_attachments = True
            data['attachments'] = attachment_fields
            
            if (files is None):
                form = None
            
            else:
                form = Formdata()
                form.add_field('payload_json', to_json(data))
                
                for index, (name, io, description) in enumerate(files):
                    form.add_field(f'files[{index}]', io, filename=name, content_type='application/octet-stream')
        
        
        return form, contains_attachments
    
    
    def add_file_to_message_data(message_data, file, contains_content, is_edit):
        """
        Adds files to the message data creating a form data if applicable.
        
        Parameters
        ----------
        message_data : `dict` of (`str`, `Any`) items
            The message's payload to send.
        file : `None`, `dict` of (`file-name`, `io`) items, `list` of (`file-name`, `io`) elements, \
                tuple (`file-name`, `io`), `io`
            The files to send.
        contains_content : `bool`
            Whether the message already contains any content.
        is_edit : `bool`
            Whether we are creating edit file form.
        
        Returns
        -------
        message_data : `None`, `dict`, ``Formdata``
            Returns a ``Formdata`` if the message contains attachments, `dict` if contains any content and `None` if
            not.
        
        Raises
        ------
        ValueError
            If more than `10` file is registered to be sent.
        """
        if (file is (... if is_edit else None)):
            if not contains_content:
                message_data = None
        else:
            form, contains_attachments = create_file_form(message_data, file)
            if (form is None):
                if (not contains_content) and (not contains_attachments):
                    message_data = None
            else:
                message_data = form
        
        return message_data

else:
    def create_file_form(data, file):
        """
        Creates a `multipart/form-data` form from the message's data and from the file data. If there is no files to
        send, will return `None` to tell the caller, that nothing is added to the overall data.
        
        Parameters
        ----------
        data : `dict` of `Any`
            The data created by the ``.message_create`` method.
        file : `dict` of (`file-name`, `io`) items, `list` of (`file-name`, `io`) elements, tuple (`file-name`, `io`),
                `io`
            The files to send.
        
        Returns
        -------
        form : `None`, `Formdata`
            Returns a `Formdata` of the files and from the message's data. If there are no files to send, returns `None`
            instead.
        
        Raises
        ------
        ValueError
            If more than `10` file is registered to be sent.
        
        Notes
        -----
        Accepted `io` types with check order are:
        - ``BodyPartReader``
        - `bytes`, `bytearray`, `memoryview`
        - `str`
        - `BytesIO`
        - `StringIO`
        - `TextIOBase`
        - `BufferedReader`, `BufferedRandom`
        - `IOBase`
        - ``AsyncIO``
        - `async-iterable`
        
        Raises `TypeError` at the case of invalid `io` type.
        
        There are two predefined data types specialized to send files:
        - ``ReuBytesIO``
        - ``ReuAsyncIO``
        
        If a buffer is sent, then when the request is done, it is closed. So if the request fails, we would not be
        able to resend the file, except if we have a data type, what instead of closing on `.close()` just seeks to
        `0` (or later if needed) on close, instead of really closing instantly. These data types implement a
        `.real_close()` method, but they do `real_close` on `__exit__` as well.
        """
        form = Formdata()
        form.add_field('payload_json', to_json(data))
        files = []
        
        # checking structure
        
        # case 0 none
        if file is None:
            pass
        
        # case 1 dict like
        elif hasattr(type(file), 'items'):
            files.extend(file.items())
        
        # case 2 tuple => file, filename pair
        elif isinstance(file, tuple):
            files.append(file)
        
        # case 3 list like
        elif isinstance(file, (list, deque)):
            for element in file:
                if type(element) is tuple:
                    name, io = element
                else:
                    io = element
                    name = ''
                
                if not name:
                    #guessing name
                    name = getattr(io, 'name', '')
                    if name:
                        _, name = split_path(name)
                    else:
                        name = str(random_id())
                
                files.append((name, io),)
        
        #case 4 file itself
        else:
            name = getattr(file, 'name', '')
            #guessing name
            if name:
                _, name = split_path(name)
            else:
                name = str(random_id())
            
            files.append((name, file),)
        
        # checking the amount of files
        # case 1 one file
        if len(files) == 1:
            name, io = files[0]
            form.add_field('file', io, filename=name, content_type='application/octet-stream')
        # case 2, no files -> return None, we should use the already existing data
        elif len(files) == 0:
            return None
        # case 3 maximum 10 files
        elif len(files) < 11:
            for index, (name, io) in enumerate(files):
                form.add_field(f'file{index}s', io, filename=name, content_type='application/octet-stream')
        
        # case 4 more than 10 files
        else:
            raise ValueError(
                f'Can send up to 10 files at once, got {len(files)!r}; {reprlib.repr(files)}.'
            )
        
        return form
    
    
    def add_file_to_message_data(message_data, file, contains_content, is_edit):
        """
        Adds files to the message data creating a form data if applicable.
        
        Parameters
        ----------
        message_data : `dict` of (`str`, `Any`) items
            The message's payload to send.
        file : `None`, `dict` of (`file-name`, `io`) items, `list` of (`file-name`, `io`) elements, \
                tuple (`file-name`, `io`), `io`
            The files to send.
        contains_content : `bool`
            Whether the message already contains any content.
        is_edit : `bool`
            Whether we are creating edit file form.
        
        Returns
        -------
        message_data : `None`, `dict`, ``Formdata``
            Returns a ``Formdata`` if the message contains attachments, `dict` if contains any content and `None` if
            not.
        
        Raises
        ------
        ValueError
            If more than `10` file is registered to be sent.
        """
        if (file is (... if is_edit else None)):
            if not contains_content:
                message_data = None
        else:
            form = create_file_form(message_data, file)
            if (form is None) and (not contains_content):
                message_data = None
            else:
                message_data = form
        
        return message_data


def validate_content_and_embed(content, embed, is_edit):
    """
    Validates the given content and embed fields of a message creation or edition.
    
    Parameters
    ----------
    content : `str`, ``EmbedBase``, `Any`, Optional
        The content of the message.
        
        
        If given as ``EmbedBase``, then the message's embeds will be edited with it.
    embed : `None`, ``EmbedBase``, `list` of ``EmbedBase``, Optional (Keyword only)
        The new embedded content of the message. By passing it as `None`, you can remove the old.
        
        > If `embed` and `content` parameters are both given as  ``EmbedBase``, then `AssertionError` is
        raised.
    is_edit : `bool`
        Whether the processed `content` and `embed` fields are for message edition. At this case passing `None` will
        remove them.
    
    Returns
    -------
    content : `Ellipsis`, `None`, `str`
        The message's content.
    embed : `Ellipsis`, `None`, ``EmbedBase``, (`list`, `tuple`) of ``EmbedBase``
        The messages embeds.
    
    Raises
    ------
    TypeError
        If `embed` was not given neither as ``EmbedBase`` nor as `list`, `tuple` of ``EmbedBase``-s.
    AssertionError
        - If `embed` contains a non ``EmbedBase`` element.
        - If both `content` and `embed` fields are embeds.
    """
    # Embed check order:
    # 1.: None
    # 2.: Ellipsis -> None || Ellipsis
    # 3.: Embed : -> embed || [embed]
    # 4.: list of Embed -> embed[0] || embed[:10] or None
    # 5.: raise
    if embed is None:
        pass
    
    elif embed is ...:
        if not is_edit:
            embed = None
    
    elif isinstance(embed, EmbedBase):
        embed = [embed]
    
    elif isinstance(embed, (list, tuple)):
        if embed:
            if __debug__:
                for embed_element in embed:
                    if not isinstance(embed_element, EmbedBase):
                        raise AssertionError(
                            f'`embed` can contains `{EmbedBase.__name__}` elements, got '
                            f'{embed_element.__class__.__name__}; {embed_element!r}; embed={embed!r}.'
                        )
            
            embed = embed[:10]
        else:
            embed = None
    
    else:
        raise TypeError(
            f'`embed` can be `{EmbedBase.__name__}`, (`list`, `tuple`) of {EmbedBase.__name__}, got '
            f'{embed.__class__.__name__}; {embed!r}.'
        )
    
    # Content check order:
    # 1.: None -> None || ''
    # 2.: Ellipsis -> None || Ellipsis
    # 3.: str
    # 4.: Embed -> embed = content || [content]
    # 5.: list of Embed -> embed = content[0] || content[:10]
    # 6.: object -> str(content)
    
    if content is None:
        if is_edit:
            content = ''
    
    elif content is ...:
        if not is_edit:
            content = None
    
    elif isinstance(content, str):
        pass
    
    elif isinstance(content, EmbedBase):
        if __debug__:
            if (embed is not (... if is_edit else None)):
                raise AssertionError(
                    f'Multiple parameters were given as embed, got content={content!r}, embed={embed!r}.'
                )
        
        embed = [content]
        
        if is_edit:
            content = ...
        else:
            content = None
    
    else:
        # Check for list of embeds as well.
        if isinstance(content, (list, tuple)):
            if content:
                for element in content:
                    if isinstance(element, EmbedBase):
                        continue
                    
                    is_list_of_embeds = False
                    break
                else:
                    is_list_of_embeds = True
            else:
                is_list_of_embeds = False
        else:
            is_list_of_embeds = False
        
        if is_list_of_embeds:
            if __debug__:
                if (embed is not (... if is_edit else None)):
                    raise AssertionError(
                        f'Multiple parameters were given as embed, got content={content!r}, embed={embed!r}.'
                    )
            
            embed = content[:10]
            
            if is_edit:
                content = ...
            else:
                content = None
        else:
            content = str(content)
    
    return content, embed


TIMEOUT_MAX_DURATION = timedelta(days=28)
TIMEOUT_MAX_DURATION_SECONDS_FLOAT = TIMEOUT_MAX_DURATION.total_seconds()
TIMEOUT_MAX_DURATION_SECONDS_INT = floor(TIMEOUT_MAX_DURATION_SECONDS_FLOAT)
ZERO_TIMEDELTA = timedelta(seconds=0)


def validate_timeout_duration(timeout_duration):
    """
    Validates the given `timeout_duration` and returns the processed value which equals to the timeout's end.
    
    Parameters
    ----------
    timeout_duration: `None`, `int`, `float`, `timedelta`, `datetime`, Optional (Keyword only)
        The timeout duration of the user in seconds.
        
        Pass it as `None` or as a non-positive duration to remove it.
        
        The max allowed value equals to 28 days.
    
    Returns
    -------
    timeout_ends_at : `None`, `datetime`
    
    Raises
    ------
    TypeError
        - If `timeout_duration`'s type is incorrect.
    """
    if timeout_duration is None:
        timeout_ends_at = None
    
    elif isinstance(timeout_duration, int):
        if timeout_duration <= 0:
            timeout_ends_at = None
        
        elif timeout_duration >= TIMEOUT_MAX_DURATION_SECONDS_INT:
            timeout_ends_at = datetime.utcnow() + TIMEOUT_MAX_DURATION
        
        else:
            timeout_ends_at = datetime.utcnow() + timedelta(seconds=timeout_duration)
    
    elif isinstance(timeout_duration, float):
        if timeout_duration <= 0.0:
            timeout_ends_at = None
        
        elif timeout_duration >= TIMEOUT_MAX_DURATION_SECONDS_FLOAT:
            timeout_ends_at = datetime.utcnow() + TIMEOUT_MAX_DURATION
        
        else:
            timeout_ends_at = datetime.utcnow() + timedelta(seconds=timeout_duration)
    
    elif isinstance(timeout_duration, timedelta):
        if timeout_duration <= ZERO_TIMEDELTA:
            timeout_ends_at = None
        
        elif timeout_duration >= TIMEOUT_MAX_DURATION:
            timeout_ends_at = datetime.utcnow() + TIMEOUT_MAX_DURATION
        
        else:
            timeout_ends_at = datetime.utcnow() + timeout_duration
    
    elif isinstance(timeout_duration, datetime):
        timeout_ends_at = timeout_duration
    
    else:
        raise TypeError(
            f'`timeout_duration` can be `None`, `int`, `float`, `timedelta`, `datetime`, got '
            f'{timeout_duration.__class__.__name__}; {timeout_duration!r}.'
        )

    return timeout_ends_at


def get_channel_id(channel, type_checker):
    """
    Gets the channel's identifier from the given channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `int`
        The channel, or it's identifier.
    type_checker : `FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    if isinstance(channel, Channel):
        if type_checker(channel) or channel.partial:
            return channel.id
    
    else:
        channel_id = maybe_snowflake(channel)
        if (channel_id is not None):
            return channel_id
        
    raise TypeError(
        f'`channel` can be `{Channel.__name__}`, `int`,  passing the `{type_checker.__name__}` check, '
        f'got {channel.__class__.__name__}; {channel!r}.'
    )


def get_channel_guild_id_and_id(channel, type_checker):
    """
    Gets the channel's and it's guild's identifier from the given channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `tuple` (`int`, `int`)
        The role, or `guild-id`, `role-id` pair.
    type_checker : `FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    snowflake_pair : `tuple` (`int`, `int`)
        The channel's guild's and it's own identifier if applicable.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    if isinstance(channel, Channel):
        if type_checker(channel) or channel.partial:
            return channel.guild_id, channel.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(channel)
        if (snowflake_pair is not None):
            return snowflake_pair
    
    raise TypeError(
        f'`channel` can be `{Channel.__name__}`, `int`,  passing the `{type_checker.__name__}` check, '
        f'got {channel.__class__.__name__}; {channel!r}.'
    )


def get_channel_and_id(channel, type_checker):
    """
    Gets thread channel and it's identifier from the given thread channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `int`
        The channel, or it's identifier.
    type_checker : `FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    channel : `None`, ``Channel``
        The channel if found.
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    while True:
        if isinstance(channel, Channel):
            if type_checker(channel) or channel.partial:
                channel_id = channel.id
                break
        
        else:
            channel_id = maybe_snowflake(channel)
            if (channel_id is not None):
                try:
                    channel = CHANNELS[channel_id]
                except KeyError:
                    channel = None
                    break
                
                if type_checker(channel) or channel.partial:
                    channel_id = channel.id
                    break
        
        raise TypeError(
            f'`channel` can be `{Channel.__name__}`, `int`,  passing the `{type_checker.__name__}` check, '
            f'got {channel.__class__.__name__}; {channel!r}.'
        )
    
    return channel, channel_id


def get_stage_channel_id(stage):
    """
    Gets stage's channel's identifier from the given stage or of it's identifier.
    
    Parameters
    ----------
    stage : `Stage`, ``Channel``, `int`
        The stage, or it's identifier.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `stage`'s type is incorrect.
    """
    while True:
        if isinstance(stage, Stage):
            channel_id = stage.channel.id
            break
        
        elif isinstance(stage, Channel):
            if stage.is_guild_stage() or stage.partial:
                channel_id = stage.id
                break
        
        else:
            channel_id = maybe_snowflake(stage)
            if (channel_id is not None):
                break
        
        raise TypeError(
            f'`stage` can be `{Stage.__name__}`, `{Channel.__name__}`, `int`, got '
            f'{stage.__class__.__name__}; {stage!r}.'
        )
    
    return channel_id


def get_user_id(user):
    """
    Gets user identifier from the given user or of it's identifier.
    
    Parameters
    ----------
    user : ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    if isinstance(user, ClientUserBase):
        return user.id
    
    user_id = maybe_snowflake(user)
    if (user_id is not None):
        return user_id
    
    raise TypeError(
        f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
    )


def get_user_and_id(user):
    """
    Gets user and it's identifier from the given user or of it's identifier.
    
    Parameters
    ----------
    user : ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user : `None`, ``ClientUserBase``
        The user if found.
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    while True:
        if isinstance(user, ClientUserBase):
            user_id = user.id
            break
        
        user_id = maybe_snowflake(user)
        if (user_id is not None):
            try:
                user = USERS[user_id]
            except KeyError:
                user = None
                break
            
            if isinstance(user, ClientUserBase):
                break
        
        raise TypeError(
            f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
        )
        
    return user, user_id


def get_user_id_nullable(user):
    """
    Gets user identifier from the given user or of it's identifier.
    
    > The user can be `None`. At that case `user_id` will default to `0`.
    
    Parameters
    ----------
    user : `None`, ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    if user is None:
        return 0
    
    if isinstance(user, ClientUserBase):
        return user.id
    
    user_id = maybe_snowflake(user)
    if (user_id is not None):
        return user_id
        
    raise TypeError(
        f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
    )


def get_guild_id(guild):
    """
    Gets the guild's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        return guild.id
    
    guild_id = maybe_snowflake(guild)
    if (guild_id is not None):
        return guild_id
        
    raise TypeError(
        f'`guild` can be `{Guild.__name__}`, `int`, got {guild.__class__.__name__}; {guild!r}.'
    )


def get_guild_and_id(guild):
    """
    Gets the guild and it's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild : `None`, ``Guild``
        The guild if found.
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        guild_id = guild.id
    else:
        guild_id = maybe_snowflake(guild)
        if guild_id is None:
            raise TypeError(
                f'`guild` can be `{Guild.__name__}`, `int`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        guild = GUILDS.get(guild_id, None)
    
    return guild, guild_id


def get_achievement_id(achievement):
    """
    Gets the achievement identifier from the given achievement or of it's identifier.
    
    Parameters
    ----------
    achievement : ``Achievement``, `int`
        The achievement, or it's identifier.
    
    Returns
    -------
    achievement_id : `int`
        The achievement's identifier.
    
    Raises
    ------
    TypeError
        If `achievement`'s type is incorrect.
    """
    if isinstance(achievement, Achievement):
        achievement_id = achievement.id
    else:
        achievement_id = maybe_snowflake(achievement)
        if achievement_id is None:
            raise TypeError(
                f'`achievement` can be `{Achievement.__name__}`, `int`, got '
                f'{achievement.__class__.__name__}; {achievement!r}.'
            )
    
    return achievement_id


def get_achievement_and_id(achievement):
    """
    Gets the achievement and it's identifier from the given achievement or of it's identifier.
    
    Parameters
    ----------
    achievement : ``Achievement``, `int`
        The achievement, or it's identifier.
    
    Returns
    -------
    achievement : ``Achievement``, `None`
        The achievement if found.
    achievement_id : `int`
        The achievement's identifier.
    
    Raises
    ------
    TypeError
        If `achievement`'s type is incorrect.
    """
    if isinstance(achievement, Achievement):
        achievement_id = achievement.id
    else:
        achievement_id = maybe_snowflake(achievement)
        if achievement_id is None:
            raise TypeError(
                f'`achievement` can be `{Achievement.__name__}`, `int`, got '
                f'{achievement.__class__.__name__}; {achievement!r}.'
            )
        
        achievement = None
    
    return achievement, achievement_id


def get_channel_id_and_message_id(message):
    """
    Gets the message's channel's and it's own identifier.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message or it' representation.
    
    Returns
    -------
    channel_id : `int`
        The message's channel's identifier.
    message_id : `int`
        The message's identifier.
    
    Raises
    ------
    TypeError
        If `message`'s type is incorrect.
    """
    # 1.: Message
    # 4.: None -> raise
    # 5.: `tuple` (`int`, `int`)
    # 6.: raise
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    elif message is None:
        raise TypeError(
            '`message` was given as `None`. Make sure to call message create methods with non-empty content(s).'
        )
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {message.__class__.__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
    
    return channel_id, message_id


def get_message_and_channel_id_and_message_id(message):
    """
    Gets the message's channel's and it's own identifier.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message or it' representation.
    
    Returns
    -------
    message : `None`, ``Message``
        The message in context if found.
    channel_id : `int`
        The message's channel's identifier.
    message_id : `int`
        The message's identifier.
    
    Raises
    ------
    TypeError
        If `message`'s type is incorrect.
    """
    # 1.: Message
    # 4.: None -> raise
    # 5.: `tuple` (`int`, `int`)
    # 6.: raise
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    elif message is None:
        raise TypeError(
            '`message` was given as `None`. Make sure to call message create methods with non-empty content(s).'
        )
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {message.__class__.__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
        message = MESSAGES.get(message_id, None)
    
    return message, channel_id, message_id


def get_role_id(role):
    """
    Gets the role identifier from the given role or of it's identifier.
    
    Parameters
    ----------
    role : ``Role``, `int`
        The role, or it's identifier.
    
    Returns
    -------
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        role_id = role.id
    else:
        role_id = maybe_snowflake(role)
        if role_id is None:
            raise TypeError(
                f'`role` can be `{Role.__name__}`, `int`, got {role.__class__.__name__}; {role!r}.'
            )
    
    return role_id


def get_role_role_guild_id_and_id(role):
    """
    Gets the role identifier from the given role or of it's identifier.
    
    Parameters
    ----------
    role : ``Role``, `tuple` (`int`, `int`)
        The role, or a `guild-id`, `role-id` pair.
    
    Returns
    -------
    role : `None`, ``Role``
        The respective role.
    
    guild_id : `int`
        The role's guild's identifier.
    
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        role_id = role.id
        guild_id = role.guild_id
    
    else:
        snowflake_pair = maybe_snowflake_pair(role)
        if snowflake_pair is None:
            raise TypeError(
                f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got {role.__class__.__name__}; {role!r}.'
            )
        
        guild_id, role_id = snowflake_pair
        role = ROLES.get(role_id, None)
    
    return role, guild_id, role_id


def get_role_guild_id_and_id(role):
    """
    Gets the role's and it's guild's identifier from the given role or of a `guild-id`, `role-id` pair.
    
    Parameters
    ----------
    role : ``Role``, `tuple` (`int`, `int`)
        The role, or a `guild-id`, `role-id` pair.
    
    Returns
    -------
    guild_id : `int`
        The role's guild's identifier.
    
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        return role.guild_id, role.id
    
    snowflake_pair = maybe_snowflake_pair(role)
    if snowflake_pair is not None:
        return snowflake_pair
    
    raise TypeError(
        f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got {role.__class__.__name__}; {role!r}.'
    )
    

def get_webhook_id(webhook):
    """
    Gets the webhook's identifier from the given webhook or of it's identifier.
    
    Parameters
    ----------
    webhook : ``Webhook``, `int`
        The webhook, or it's identifier.
    
    Returns
    -------
    webhook_id : `int`
        The webhook's identifier.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    if isinstance(webhook, Webhook):
        webhook_id = webhook.id
    else:
        webhook_id = maybe_snowflake(webhook)
        if webhook_id is None:
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `int`, got {webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return webhook_id


def get_webhook_and_id(webhook):
    """
    Gets the webhook and it's identifier from the given webhook or of it's identifier.
    
    Parameters
    ----------
    webhook : ``Webhook``, `int`
        The webhook, or it's identifier.
    
    Returns
    -------
    webhook : ``Webhook``, `None`
        The webhook if any.
    webhook_id : `int`
        The webhook's identifier.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    while True:
        if isinstance(webhook, Webhook):
            webhook_id = webhook.id
            break
        
        webhook_id = maybe_snowflake(webhook)
        if (webhook_id is not None):
            try:
                webhook = USERS[webhook_id]
            except KeyError:
                webhook = None
                break
            
            if isinstance(webhook, Webhook):
                break
            
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `int`, got {webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return webhook, webhook_id


def get_webhook_id_and_token(webhook):
    """
    Gets the webhook's identifier and token from the given webhook or it's token, identifier pair.
    
    Parameters
    ----------
    webhook : ``Webhook``, `tuple` (`int`, `str`)
        The webhook or it's id and token.
    
    Returns
    -------
    webhook_id : `int`
        The webhook's identifier.
    webhook_token : `str`
        The webhook's token.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    if isinstance(webhook, Webhook):
        snowflake_token_pair = webhook.id, webhook.token
    else:
        snowflake_token_pair = maybe_snowflake_token_pair(webhook)
        if (snowflake_token_pair is None):
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `tuple` (`int`, `str`), got '
                f'{webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return snowflake_token_pair


def get_webhook_and_id_and_token(webhook):
    """
    Gets the webhook, it's identifier and token from the given webhook or it's token, identifier pair.
    
    Parameters
    ----------
    webhook : ``Webhook``, `tuple` (`int`, `str`)
        The webhook or it's id and token.
    
    Returns
    -------
    webhook : ``Webhook``, `None`
        The webhook if any.
    webhook_id : `int`
        The webhook's identifier.
    webhook_token : `str`
        The webhook's token.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    while True:
        if isinstance(webhook, Webhook):
            webhook_id = webhook.id
            webhook_token = webhook.token
            break
        
        snowflake_token_pair = maybe_snowflake_token_pair(webhook)
        if (snowflake_token_pair is not None):
            webhook_id, webhook_token = snowflake_token_pair
            
            try:
                webhook = USERS[webhook_id]
            except KeyError:
                webhook = None
                break
            
            if isinstance(webhook, Webhook):
                break
        
        raise TypeError(
            f'`webhook` can be `{Webhook.__name__}`, `tuple` (`int`, `str`), got '
            f'{webhook.__class__.__name__}; {webhook!r}.'
        )
    
    return webhook, webhook_id, webhook_token


def get_reaction(emoji):
    """
    Gets the reaction form of the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji``, `str`
        The emoji to get it's reaction form of.
    
    Returns
    -------
    as_reaction : `str`
        The emoji's reaction form.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    """
    if isinstance(emoji, Emoji):
        as_reaction = emoji.as_reaction
    elif isinstance(emoji, str):
        as_reaction = emoji
    else:
        raise TypeError(
            f'`emoji` can be `{Emoji.__class__}`, `str`, got {emoji.__class__.__name__}; {emoji!r}.'
        )
    
    return as_reaction


def get_emoji_from_reaction(emoji):
    """
    Gets emoji from the given emoji or reaction string.
    
    Parameters
    ----------
    emoji : ``Emoji``, `str`
        The emoji or reaction to get the emoji from.
    
    Returns
    -------
    emoji : ``Emoji``
        The emoji itself.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    ValueError
        The given `emoji` is not a valid reaction.
    """
    if isinstance(emoji, Emoji):
        pass
    elif isinstance(emoji, str):
        emoji = parse_reaction(emoji)
        if emoji is None:
            raise ValueError(
                f'The given `emoji` is not a valid reaction, got {emoji!r}.'
            )
    else:
        raise TypeError(
            f'`emoji` can be `{Emoji.__class__}`, `str`, got {emoji.__class__.__name__}; {emoji!r}.'
        )
    
    return emoji


def get_emoji_guild_id_and_id(emoji):
    """
    Gets the emoji's and it's guild's identifier from the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji``, `tuple` (`int`, `int`)
        The emoji, or `guild-id`, `emoji-id` pair.
    
    Returns
    -------
    snowflake_pair : `tuple` (`int`, `int`)
        The emoji's guild's and it's own identifier if applicable.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    """
    if isinstance(emoji, Emoji):
        snowflake_pair = emoji.guild_id, emoji.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(emoji)
        if snowflake_pair is None:
            raise TypeError(
                f'`emoji` can be `{Emoji.__name__}`, `tuple` (`int`, `int`), got {emoji.__class__.__name__}; '
                f'{emoji!r}.'
            )
    
    return snowflake_pair


def get_sticker_and_id(sticker):
    """
    Gets sticker and it's identifier from the given sticker or of it's identifier.
    
    Parameters
    ----------
    sticker : ``Sticker``, `int`
        The sticker, or it's identifier.
    
    Returns
    -------
    sticker : ``Sticker``, `None`
        The sticker if found.
    sticker_id : `int`
        The sticker's identifier.
    
    Raises
    ------
    TypeError
        If `sticker`'s type is incorrect.
    """
    while True:
        if isinstance(sticker, Sticker):
            sticker_id = sticker.id
            break
        
        sticker_id = maybe_snowflake(sticker)
        if (sticker_id is not None):
            sticker = STICKERS.get(sticker_id, None)
            break
        
        raise TypeError(
            f'`sticker` can be `{Sticker.__name__}`, `int`, got {sticker.__class__.__name__}; {sticker!r}.'
        )
        
    return sticker, sticker_id


def get_sticker_pack_and_id(sticker_pack):
    """
    Gets sticker pack and it's identifier from the given sticker pack or of it's identifier.
    
    Parameters
    ----------
    sticker_pack : ``StickerPack``, `int`
        The sticker, or it's identifier.
    
    Returns
    -------
    sticker_pack : ``StickerPack``, `None`
        The sticker pack if found.
    sticker_pack_id : `int`
        The sticker pack's identifier.
    
    Raises
    ------
    TypeError
        If `sticker_pack`'s type is incorrect.
    """
    while True:
        if isinstance(sticker_pack, StickerPack):
            sticker_pack_id = sticker_pack.id
            break
        
        sticker_pack_id = maybe_snowflake(sticker_pack)
        if (sticker_pack_id is not None):
            sticker_pack = STICKER_PACKS.get(sticker_pack_id, None)
            break
        
        raise TypeError(
            f'`sticker_pack` can be `{StickerPack.__name__}`, `int`, got {sticker_pack.__class__.__name__}; '
            f'{sticker_pack!r}.'
        )
        
    return sticker_pack, sticker_pack_id


def get_scheduled_event_guild_id_and_id(scheduled_event):
    """
    Gets the scheduled event's and it's identifier from the given scheduled event or from a tuple of 2 identifiers.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``, `tuple` of (`int`, `int`)
        The scheduled event, or a tuple of two identifiers.
    
    Returns
    -------
    guild_id : `int`
        The scheduled event's guild's identifier.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Raises
    ------
    TypeError
        If `scheduled_event`'s type is incorrect.
    """
    if isinstance(scheduled_event, ScheduledEvent):
        return scheduled_event.guild_id, scheduled_event.id
    
    
    snowflake_pair = maybe_snowflake_pair(scheduled_event)
    if snowflake_pair is None:
        raise TypeError(
            f'`scheduled_event` can be `{ScheduledEvent.__name__}`, `tuple` of (`int`, `int`), got '
            f'{scheduled_event.__class__.__name__}; {scheduled_event!r}.'
        )
    
    return snowflake_pair


def get_scheduled_event_and_guild_id_and_id(scheduled_event):
    """
    Gets the scheduled event, it's guild's identifier and it's identifier.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``, `tuple` (`int`, `int`)
        The scheduled event, or it's identifier.
    
    Returns
    -------
    scheduled_event : `None`, ``ScheduledEvent``
        The scheduled event.
    guild_id : `int`
        The scheduled event's guild's identifier.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Raises
    ------
    TypeError
        If `scheduled_event`'s type is incorrect.
    """
    if isinstance(scheduled_event, ScheduledEvent):
        scheduled_event =  scheduled_event
        scheduled_event_id = scheduled_event.id
        guild_id = scheduled_event.guild_id
    else:
        snowflake_pair = maybe_snowflake_pair(scheduled_event)
        if snowflake_pair is None:
            raise TypeError(
                f'`scheduled_event` can be `{ScheduledEvent.__name__}`, `tuple` of (`int`, `int`), got '
                f'{scheduled_event.__class__.__name__}; {scheduled_event!r}.'
            )
        
        guild_id, scheduled_event_id = snowflake_pair
        
        scheduled_event = SCHEDULED_EVENTS.get(scheduled_event_id, None)
    
    return scheduled_event, guild_id, scheduled_event_id


def get_application_command_id(application_command):
    """
    Gets the application command's identifier.
    
    Parameters
    ----------
    application_command : ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
    
    return application_command_id


def get_application_command_and_id(application_command):
    """
    Gets the application command's identifier.
    
    Parameters
    ----------
    application_command : ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command : `None`, ``ApplicationCommand``
        The application command if exists.
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
        
        application_command = APPLICATION_COMMANDS.get(application_command_id, None)
    
    return application_command, application_command_id


def get_application_command_id_nullable(application_command):
    """
    Gets the application command's identifier.
    
    > The application command can be `None`. At that case `application_command_id` will default to `0`.
    
    Parameters
    ----------
    application_command : `None`, ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if application_command is None:
        application_command_id = 0
    
    elif isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
    
    return application_command_id


def get_auto_moderation_rule_guild_id_and_id(auto_moderation_rule):
    """
    Gets the auto moderation rule's guild's identifier and it's own identifier.
    
    Parameters
    ----------
    auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
        The auto moderation rule, or it's identifier as a `guild-id`, `rule-id` pair..
    
    Returns
    -------
    guild_id : `int`
        The auto moderation rule's guild's identifier.
    auto_moderation_rule_id : `int`
        The auto moderation rule's identifier.
    
    Raises
    ------
    TypeError
        If `auto_moderation_rule`'s type is incorrect.
    """
    if isinstance(auto_moderation_rule, AutoModerationRule):
        guild_id = auto_moderation_rule.guild_id
        rule_id = auto_moderation_rule.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(auto_moderation_rule)
        if snowflake_pair is None:
            raise TypeError(
                f'`auto_moderation_rule` can be `{AutoModerationRule.__name__}`, `tuple` of (`int`, `int`), got '
                f'{auto_moderation_rule.__class__.__name__}; {auto_moderation_rule!r}.'
            )
        
        guild_id, rule_id = snowflake_pair
    
    return guild_id, rule_id


def get_auto_moderation_rule_and_guild_id_and_id(auto_moderation_rule):
    """
    Gets the auto moderation, its guild's identifier and it's own identifier.
    
    Parameters
    ----------
    auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
        The auto moderation rule, or it's identifier as a `guild-id`, `rule-id` pair..
    
    Returns
    -------
    auto_moderation_rule : `None`, ``AutoModerationRule``
        The auto moderation rule.
    guild_id : `int`
        The auto moderation rule's guild's identifier.
    auto_moderation_rule_id : `int`
        The auto moderation rule's identifier.
    
    Raises
    ------
    TypeError
        If `auto_moderation_rule`'s type is incorrect.
    """
    if isinstance(auto_moderation_rule, AutoModerationRule):
        guild_id = auto_moderation_rule.guild_id
        rule_id = auto_moderation_rule.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(auto_moderation_rule)
        if snowflake_pair is None:
            raise TypeError(
                f'`auto_moderation_rule` can be `{AutoModerationRule.__name__}`, `tuple` of (`int`, `int`), got '
                f'{auto_moderation_rule.__class__.__name__}; {auto_moderation_rule!r}.'
            )
        
        guild_id, rule_id = snowflake_pair
        auto_moderation_rule = AUTO_MODERATION_RULES.get(rule_id, None)
    
    return auto_moderation_rule, guild_id, rule_id


def get_forum_tag_id(forum_tag):
    """
    Gets the forum tag's identifier.
    
    Parameters
    ----------
    forum_tag : ``ForumTag``, `int`
        The forum tag or its identifier.
    
    Returns
    -------
    forum_tag_id : `int`
        The forum tag's identifier.
    
    Raises
    ------
    TypeError
        - If `forum_tag` type is incorrect.
    """
    if isinstance(forum_tag, ForumTag):
        forum_tag_id = forum_tag.id
    
    else:
        forum_tag_id = maybe_snowflake(forum_tag)
        if (forum_tag_id is None):
            raise TypeError(
                f'`forum_tag` can be `{ForumTag.__name__}, `int`, got {forum_tag.__class__.__name__}; {forum_tag!r}.'
            )
    
    return forum_tag_id


def get_forum_tag_and_id(forum_tag):
    """
    Gets the forum tag and it's identifier.
    
    Parameters
    ----------
    forum_tag : ``ForumTag``, `int`
        The forum tag or its identifier.
    
    Returns
    -------
    forum_tag : `None`, ``ForumTag``
        The forum tag if exists.
    forum_tag_id : `int`
        The forum tag's identifier.
    
    Raises
    ------
    TypeError
        - If `forum_tag` type is incorrect.
    """
    if isinstance(forum_tag, ForumTag):
        forum_tag_id = forum_tag.id
    
    else:
        forum_tag_id = maybe_snowflake(forum_tag)
        if (forum_tag_id is None):
            raise TypeError(
                f'`forum_tag` can be `{ForumTag.__name__}, `int`, got {forum_tag.__class__.__name__}; {forum_tag!r}.'
            )
        
        forum_tag = FORUM_TAGS.get(forum_tag_id, None)
    
    return forum_tag, forum_tag_id


def get_permission_overwrite_target_id(permission_overwrite):
    """
    Returns the given `permission_overwrite`'s target id.
    
    Parameters
    ----------
    permission_overwrite : ``PermissionOverwrite``
        The permission overwrite or it's target id.
    
    Returns
    -------
    target_id : `int`
    
    Raises
    ------
    TypeError
        - If `permission_overwrite` type is incorrect.
    """
    if isinstance(permission_overwrite, PermissionOverwrite):
        target_id = permission_overwrite.target_id
    
    else:
        target_id = maybe_snowflake(permission_overwrite)
        if (target_id is None):
            raise TypeError(
                f'`permission_overwrite` can be `{PermissionOverwrite.__name__}`, `int` got '
                f'{permission_overwrite.__class__.__name__}; {permission_overwrite!r}.'
            )
    
    return target_id


def _check_required_scope(access, required_scope):
    if (required_scope is not None) and (not access.has_scope(required_scope)):
        raise ValueError(
            f'The given `access` not grants {required_scope.name!r} scope, what is required, '
            f'got {access!r}.'
        )


def get_oauth2_access_token(access, required_scope = None):
    """
    Returns the given oauth2 access's access token. Accepts both access and just token as well.
    
    Parameters
    ----------
    access : ``Oauth2Access``, ``Oauth2User``, `str`
        Oauth2 access to the respective user or it's access token.
    required_scope : `None`, ``Oauth2Scope`` = `None`, Optional
        Required scope of the access if any.
    
    Returns
    -------
    access_token : `str`
    
    Raises
    ------
    TypeError
        - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
    ValueError
        - If the given `access` is not providing the required scope.
    """
    if isinstance(access, (Oauth2Access, Oauth2User)):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
    
    elif isinstance(access, str):
        access_token = access
    
    else:
        raise TypeError(
            f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}`, `str`'
            f', got {access.__class__.__name__}; {access!r}.'
        )
    
    return access_token


def get_oauth2_access_token_and_user_id(access, user, required_scope = None):
    """
    Returns the given oauth2 access's access token and the user's identifier.
    
    Parameters
    ----------
    access : ``Oauth2Access``, ``Oauth2User``, `str`
        Oauth2 access to the respective user or it's access token.
    required_scope : `None`, ``Oauth2Scope`` = `None`, Optional
        Required scope of the access if any.
    
    Raises
    ------
    TypeError
        - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
        - If `user` is not `None`, ``ClientUserBase``.
        - If `user.id` could nto be determined.
    ValueError
        - If the given `access` is not providing the required scope.
        - If `user` and `access` refers to a different user.
    """
    user_id = get_user_id_nullable(user)
    
    if isinstance(access, Oauth2Access):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
    
    elif isinstance(access, Oauth2User):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
        
        if user_id and (user_id != access.id):
            raise ValueError(
                f'The given `user` and `access` refers to different users, got user={user!r}, '
                f'access={access!r}.'
            )
        
        user_id = access.id
    
    elif isinstance(access, str):
        access_token = access
    
    else:
        raise TypeError(
            f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}`, `str`, got '
            f'{access.__class__.__name__}; {access!r}.'
        )
    
    
    if not user_id:
        raise TypeError(
            f'`user` was not detectable neither from `user` nor from `access` parameters, got '
            f'user={user!r}, access={access!r}.'
        )
    
    return access_token, user_id



__all__ = ()

from ...field_parsers import nullable_string_parser_factory, preinstanced_parser_factory
from ...field_putters import nullable_string_putter_factory, preinstanced_putter_factory
from ...field_validators import nullable_string_validator_factory, preinstanced_validator_factory

from .preinstanced import MessageActivityType

# party_id

parse_party_id = nullable_string_parser_factory('party_id')
put_party_id_into = nullable_string_putter_factory('party_id')
validate_party_id = nullable_string_validator_factory('party_id', 0, 1024)

# type

parse_type = preinstanced_parser_factory('type', MessageActivityType, MessageActivityType.none)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', MessageActivityType)



__all__ = ('MessageActivity', )

from scarletio import RichAttributeErrorBaseType

from .fields import parse_party_id, parse_type, put_party_id_into, put_type_into, validate_party_id, validate_type
from .preinstanced import MessageActivityType


class MessageActivity(RichAttributeErrorBaseType):
    """
    Might be sent with a ``Message``, if it has rich presence-related chat embeds.
    
    Attributes
    ----------
    party_id : `None`, `str`
        The message activity's party's id.
    type : ``MessageActivityType``
        The message activity's type.
    """
    __slots__ = ('party_id', 'type',)
    
    
    def __new__(cls, *, message_activity_type = ..., party_id = ...):
        """
        Creates a new message activity from the given parameters.
        
        Parameters
        ----------
        message_activity_type : ``MessageActivityType``, `int`, Optional (Keyword only)
            The message activity's type.
        party_id : `None`, `str`, Optional (Keyword only)
            The message activity's party's id.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # message_activity_type
        if message_activity_type is ...:
            message_activity_type = MessageActivityType.none
        else:
            message_activity_type = validate_type(message_activity_type)
        
        # party_id
        if party_id is ...:
            party_id = None
        else:
            party_id = validate_party_id(party_id)
        
        self = object.__new__(cls)
        self.party_id = party_id
        self.type = message_activity_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``MessageActivity`` from message activity data included inside of a ``Message``'s data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Message activity data.
        """
        self = object.__new__(cls)
        self.party_id = parse_party_id(data)
        self.type = parse_type(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the message activity back to json a serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`)
        """
        data = {}
        put_party_id_into(self.party_id, data, defaults)
        put_type_into(self.type, data, defaults)
        return data
    
    
    def __eq__(self, other):
        """Returns whether the two message activities are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # party_id
        if self.party_id != other.party_id:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the message activity's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        party_type = self.type
        repr_parts.append(' type = ')
        repr_parts.append(repr(party_type.name))
        repr_parts.append(' ~ ')
        repr_parts.append(repr(party_type.value))
        
        party_id = self.party_id
        if (party_id is not None):
            repr_parts.append(', party_id = ')
            repr_parts.append(repr(party_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the message activity's hash value."""
        hash_value = 0
        
        # party_id
        party_id = self.party_id
        if (party_id is not None):
            hash_value ^= hash(party_id)
        
        # type
        hash_value ^= self.type.value
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the message activity.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.party_id = self.party_id
        new.type = self.type
        return new
    
    
    def copy_with(self, *, message_activity_type = ..., party_id = ...):
        """
        Copies the message activity with the given fields.
        
        Parameters
        ----------
        message_activity_type : ``MessageActivityType``, `int`, Optional (Keyword only)
            The message activity's type.
        party_id : `None`, `str`, Optional (Keyword only)
            The message activity's party's id.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # message_activity_type
        if message_activity_type is ...:
            message_activity_type = self.type
        else:
            message_activity_type = validate_type(message_activity_type)
        
        # party_id
        if party_id is ...:
            party_id = self.party_id
        else:
            party_id = validate_party_id(party_id)
        
        new = object.__new__(type(self))
        new.party_id = party_id
        new.type = message_activity_type
        return new



__all__ = ()


from .shared import convert_snowflake


def convert_application_command_id(name, data):
    return convert_snowflake('application_command_id', data)


APPLICATION_COMMAND_CONVERTERS = {
    'command_id': convert_application_command_id,
}



__all__ = (
    'create_partial_emoji_data', 'create_partial_emoji_from_data', 'create_partial_emoji_from_id',
    'create_emoji_from_exclusive_data', 'create_unicode_emoji', 'parse_all_emojis', 'parse_all_emojis_ordered',
    'parse_custom_emojis', 'parse_custom_emojis_ordered', 'parse_emoji', 'parse_reaction',
    'put_exclusive_emoji_data_into', 'put_partial_emoji_data_into', 'merge_update_reaction_mapping'
)

import warnings

from scarletio import export

from ..core import BUILTIN_EMOJIS, EMOJIS, UNICODE_TO_EMOJI
from ..utils import EMOJI_RP, REACTION_RP

from .emoji import Emoji
from .emoji_all_pattern import EMOJI_ALL_RP
from .unicode_type import Unicode, VARIATION_SELECTOR_16_POSTFIX


VARIATION_SELECTOR_16_POSTFIX_WITH_COLON = VARIATION_SELECTOR_16_POSTFIX + ':'


def _create_new_unicode(unicode_string):
    """
    Creates a new emoji from the given `unicode_string`.
    
    Parameters
    ----------
    unicode_string : `str`
        Unicode string to create emoji from.
    
    Returns
    -------
    emoji : ``Emoji``
    """
    unicode_bytes = unicode_string.encode()
    
    warnings.warn(
        f'\nUndefined emoji : {unicode_bytes!r}\nPlease open an issue with this message.',
        RuntimeWarning,
    )
    
    return Emoji._create_unicode(Unicode('', unicode_bytes, False, None, None), False)


@export
def create_partial_emoji_from_data(data):
    """
    Creates an emoji from partial emoji data sent by Discord.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Partial emoji data.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    try:
        emoji_name = data['emoji_name']
    except KeyError:
        emoji_name = data['name']
        emoji_id = data.get('id', None)
        emoji_animated = data.get('animated', False)
    else:
        emoji_id = data.get('emoji_id', None)
        emoji_animated = data.get('emoji_animated', False)
    
    if emoji_name is None:
        return None
    
    if emoji_id is None:
        try:
            emoji = UNICODE_TO_EMOJI[emoji_name]
        except KeyError:
            emoji = _create_new_unicode(emoji_name)
    
    else:
        # name can change
        if emoji_name is None:
            emoji_name = ''
            
        emoji = Emoji._create_partial(int(emoji_id), emoji_name, emoji_animated)
    
    return emoji


def create_partial_emoji_from_id(emoji_id):
    """
    Creates a partial emoji from the given id.
    
    Parameters
    ----------
    emoji_id : `int`
        The emoji's identifier.
    
    Returns
    -------
    emoji : ``Emoji``
    """
    try:
        emoji = EMOJIS[emoji_id]
    except KeyError:
        emoji = Emoji._create_empty(emoji_id)
        EMOJIS[emoji_id] = emoji
    
    return emoji


@export
def create_partial_emoji_data(emoji):
    """
    Creates partial emoji data form the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji``
        The emoji to serialize.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
        The serialized emoji data.
    """
    emoji_data = {}
    unicode = emoji.unicode
    if unicode is None:
        emoji_data['id'] = str(emoji.id)
        emoji_data['name'] = emoji.name
        
        if emoji.animated:
            emoji_data['animated'] = True
    else:
        emoji_data['name'] = unicode
    
    return emoji_data


def put_partial_emoji_data_into(emoji, data):
    """
    Familiar to ``create_partial_emoji_data``, but instead of creating a standalone emoji data, uses the `emoji_`
    prefix field form to add it to an already defined dictionary.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The emoji to serialize.
    data : `dict` of (`str`, `Any`) items
        The data to put the emoji fields into.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    if (emoji is None):
        # Require at least the `emoji_name` field.
        data['emoji_name'] = None
    
    else:
        unicode = emoji.unicode
        if unicode is None:
            data['emoji_id'] = str(emoji.id)
            data['emoji_name'] = emoji.name
            
            if emoji.animated:
                data['emoji_animated'] = True
        else:
            data['emoji_name'] = unicode
        
    return data


def create_emoji_from_exclusive_data(data):
    """
    Creates partial emoji from the given exclusive emoji data.
    
    Exclusive means, that we expect exactly `1` emoji field to be set.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Partial emoji data.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    emoji_name = data.get('emoji_name', None)
    emoji_id = data.get('emoji_id', None)
    
    if (emoji_name is not None):
        try:
            emoji = UNICODE_TO_EMOJI[emoji_name]
        except KeyError:
            emoji = _create_new_unicode(emoji_name)
    
    elif (emoji_id is not None):
        emoji = create_partial_emoji_from_id(int(emoji_id))
    
    else:
        emoji = None
    
    return emoji


def put_exclusive_emoji_data_into(emoji, data):
    """
    The reversed function of ``create_emoji_from_exclusive_data``.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The emoji to serialize.
    data : `dict` of (`str`, `Any`) items
        The data to put the emoji fields into.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    if (emoji is None):
        # Require at least the `emoji_name` field.
        data['emoji_name'] = None
    
    else:
        unicode = emoji.unicode
        if unicode is None:
            data['emoji_id'] = str(emoji.id)
        
        else:
            data['emoji_name'] = unicode
    
    return data


def parse_emoji(text):
    """
    Tries to parse out an ``Emoji`` from the inputted text. This emoji can be custom and unicode emoji as well.
    
    If the parsing fails the function returns `None`.
    
    Parameters
    ----------
    text : `str`
        The text to parse emojis from.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    parsed = EMOJI_RP.fullmatch(text)
    if (parsed is not None):
        animated, name, emoji_id = parsed.groups()
        animated = (animated is not None)
        emoji_id = int(emoji_id)
        return Emoji._create_partial(emoji_id, name, animated)
        
    try:
        return UNICODE_TO_EMOJI[text]
    except KeyError:
        pass
    
    if text.startswith(':') and text.endswith(':') and not text.endswith(VARIATION_SELECTOR_16_POSTFIX_WITH_COLON):
        try:
            return BUILTIN_EMOJIS[text[1:-1]]
        except KeyError:
            pass
    
    return None


def _iter_parse_custom_emojis(text):
    """
    Iterates over all the custom emojis in the text as they appear.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    text : `str`
        The text to parse.
    
    Yields
    ------
    emoji : ``Emoji``
    """
    for groups in EMOJI_RP.findall(text):
        
        animated, name, emoji_id = groups
        animated = (True if animated else False)
        emoji_id = int(emoji_id)
        
        yield Emoji._create_partial(emoji_id, name, animated)


def _iter_parse_all_emojis(text):
    """
    Iterates over all emojis in the text as they appear.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    text : `str`
        The text to parse.
    
    Yields
    ------
    emoji : ``Emoji``
    """
    for groups in EMOJI_ALL_RP.findall(text):
        
        unicode_value, unicode_name, custom_animated, custom_name, custom_emoji_id = groups
        if unicode_value:
            yield UNICODE_TO_EMOJI[unicode_value]
            continue
        
        if unicode_name:
            yield BUILTIN_EMOJIS[unicode_name]
            continue
        
        yield Emoji._create_partial(
            int(custom_emoji_id),
            custom_name,
            (True if custom_animated else False),
        )
        continue


def parse_custom_emojis(text):
    """
    Parses out every custom emoji from the given text.
    
    Parameters
    ----------
    text : `str`
        The text to parse.
    
    Returns
    -------
    emojis : `set` of ``Emoji``
    """
    if text is None:
        return set()
    
    return {*_iter_parse_custom_emojis(text)}


def parse_all_emojis(text):
    """
    Parses out every emoji from the given text.
    
    Parameters
    ----------
    text : `str`
        The text to parse.
    
    Returns
    -------
    emojis : `set` of ``Emoji``
    """
    if text is None:
        return set()
    return {*_iter_parse_all_emojis(text)}
    


def _parse_emojis_ordered(text, parser):
    """
    Parses emojis of the given `text` with the given `parser`.
    Returns them ordered based on their appearance in the text.
    
    
    Parameters
    ----------
    text : `None`, `str`
        The text to parse.
    parser : `GeneratorFunction`
        The parser to use.
    
    Returns
    -------
    emojis_ordered : `list` of ``Emoji``
        Excludes duplicates.
    """
    emojis_ordered = []
    if (text is not None):
        emojis_unique = set()
        
        for emoji in parser(text):
            if emoji not in emojis_unique:
                emojis_ordered.append(emoji)
                emojis_unique.add(emoji)
    
    return emojis_ordered


def parse_custom_emojis_ordered(text):
    """
    Parses out every custom emoji from the given text. Returns them ordered based on their appearance in the text.
    
    Parameters
    ----------
    text : `None`, `str`
        The text to parse.
    
    Returns
    -------
    emojis_ordered : `list` of ``Emoji``
        Excludes duplicates.
    """
    return _parse_emojis_ordered(text, _iter_parse_custom_emojis)


def parse_all_emojis_ordered(text):
    """
    Parses out every emoji from the given text. Returns them ordered based on their appearance in the text.
    
    Parameters
    ----------
    text : `None`, `str`
        The text to parse.
    
    Returns
    -------
    emojis_ordered : `list` of ``Emoji``
        Excludes duplicates.
    """
    return _parse_emojis_ordered(text, _iter_parse_all_emojis)


def parse_reaction(text):
    """
    Parses out an emoji from the given reaction string.
    
    Parameters
    ----------
    text : `str`
        Reaction string.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    try:
        emoji = UNICODE_TO_EMOJI[text]
    except KeyError:
        parsed = REACTION_RP.fullmatch(text)
        if parsed is None:
            emoji = None
        else:
            name, emoji_id = parsed.parsed()
            emoji_id = int(emoji_id)
            emoji = Emoji._create_partial(emoji_id, name, False)
    
    return emoji


@export
def create_unicode_emoji(unicode):
    """
    Creates an emoji from the given unicode value.
    
    Parameters
    ----------
    unicode : `str`
        Unicode value.
    
    Returns
    -------
    emoji : ``Emoji``
    """
    try:
        unicode_emoji = UNICODE_TO_EMOJI[unicode]
    except KeyError:
        warnings.warn(
            f'Undefined emoji : {unicode.encode()!r}\nPlease open an issue with this message.',
            RuntimeWarning,
        )
        unicode_emoji = Emoji._create_unicode(Unicode('', unicode, False, None, None), False)
    
    return unicode_emoji


def merge_update_reaction_mapping(old_reactions, new_reactions):
    """
    Merges the two reaction mapping, so values wont be overwritten if not required.
    
    Parameters
    ----------
    old_reactions : `None`, ``ReactionMapping``
        The old reactions on a message.
    new_reactions : `None`, ``ReactionMapping``
        The new reactions on a message.
    
    Returns
    -------
    real_reactions : `None`, ``ReactionMapping``
        The real merged reactions on a message.
    """
    if (old_reactions is None):
        return new_reactions
    
    elif (new_reactions is None):
        old_reactions.clear()
        return old_reactions
    
    old_emojis = {*old_reactions.keys()}
    new_emojis = {*new_reactions.keys()}
    
    for emoji in old_emojis - new_emojis:
        del old_reactions[emoji]
    
    for emoji in new_emojis - old_emojis:
        old_reactions[emoji] = new_reactions[emoji].copy()
    
    for emoji in new_emojis & old_emojis:
        old_users = old_reactions[emoji]
        new_users = new_reactions[emoji]
        
        if (len(old_users) != len(new_users)) or (not (old_users >= new_users)):
            old_reactions[emoji] = new_reactions[emoji].copy()
    
    old_reactions._full_check()
    
    return old_reactions



__all__ = ('MessageActivityType',)

from ...bases import Preinstance as P, PreinstancedBase


class MessageActivityType(PreinstancedBase):
    """
    Represents a ``MessageActivity``'s type.
    
    Attributes
    ----------
    name : `str`
        The name of the message activity type.
    value : `int`
        The Discord side identifier value of the message activity type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``MessageActivityType``) items
        Stores the predefined ``MessageActivityType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The message activity types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the message activity types.
    
    Every predefined message activity type can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | join                  | join          | 1     |
    +-----------------------+---------------+-------+
    | spectate              | spectate      | 2     |
    +-----------------------+---------------+-------+
    | listen                | listen        | 3     |
    +-----------------------+---------------+-------+
    | watch                 | watch         | 4     |
    +-----------------------+---------------+-------+
    | join_request          | join_request  | 5     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    join = P(1, 'join')
    spectate = P(2, 'spectate')
    listen = P(3, 'listen')
    watch = P(4, 'watch')
    join_request = P(5, 'join_request')



__all__ = ()

from scarletio import include

from ...field_parsers import bool_parser_factory, force_string_parser_factory, preinstanced_parser_factory
from ...field_putters import (
    bool_optional_putter_factory, force_string_putter_factory, preinstanced_optional_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import bool_validator_factory, force_string_validator_factory, preinstanced_validator_factory

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN
from .preinstanced import ConnectionType, ConnectionVisibility


Integration = include('Integration')

# parse_friend_sync

parse_friend_sync = bool_parser_factory('friend_sync', False)
put_friend_sync_into = bool_optional_putter_factory('friend_sync', False)
validate_friend_sync = bool_validator_factory('friend_sync')

# integrations

def parse_integrations(data):
    """
    Parses out an integration array from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Entity data.
    
    Returns
    -------
    entity_array : `None`, `tuple` of ``Integration``
    """
    entity_data_array = data.get('integrations', None)
    if (entity_data_array is None) or (not entity_data_array):
        entity_array = None
    else:
        entity_array = tuple(sorted(Integration.from_data(entity_data) for entity_data in entity_data_array))
    
    return entity_array


def put_integrations_into(entity_array, data, defaults, *, include_internals = False):
    """
    Puts the given integration array into the given `data` json serializable object.
    
    Parameters
    ----------
    entity_array : `None`, `tuple` of ``Integration``
        Integration array.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    include_internals : `bool` = `False`, Optional (Keyword only)
        Whether internal fields should be included.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    if entity_array is None:
        entity_data_array = []
    else:
        entity_data_array = [
            entity.to_data(defaults = defaults, include_internals = include_internals)
            for entity in entity_array
        ]
    
    data['integrations'] = entity_data_array
    
    return data


def validate_integrations(entity_array):
    """
    Validates the given nullable entity array field.
    
    Parameters
    ----------
    entity_array : `None`, `iterable` of ``Integration``
        The entity array to validate.
    
    Returns
    -------
    entity_array : `None`, `tuple` of ``Integration``
    
    Raises
    ------
    TypeError
        - If `entity_array` is not `None`, `iterable` of ``Integration``.
    """
    if entity_array is None:
        return None
    
    if (getattr(entity_array, '__iter__', None) is None):
        raise TypeError(
            f'`integrations` can be `None`, `iterable` of `{Integration.__name__}`, got '
            f'{entity_array.__class__.__name__}; {entity_array!r}.'
        )
        
    entity_array_processed = None
    
    for entity in entity_array:
        if not isinstance(entity, Integration):
            raise TypeError(
                f'`integrations` can contain `{Integration.__name__}` elements, got '
                f'{entity.__class__.__name__}; {entity!r}; entity_array = {entity_array!r}.'
            )
        
        if (entity_array_processed is None):
            entity_array_processed = set()
        
        entity_array_processed.add(entity)
    
    if (entity_array_processed is not None):
        entity_array_processed = tuple(sorted(entity_array_processed))
    
    return entity_array_processed

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# revoked

parse_revoked = bool_parser_factory('revoked', False)
put_revoked_into = bool_optional_putter_factory('revoked', False)
validate_revoked = bool_validator_factory('revoked')

# show_activity

parse_show_activity = bool_parser_factory('show_activity', False)
put_show_activity_into = bool_optional_putter_factory('show_activity', False)
validate_show_activity = bool_validator_factory('show_activity')

# two_way_link

parse_two_way_link = bool_parser_factory('two_way_link', False)
put_two_way_link_into = bool_optional_putter_factory('two_way_link', False)
validate_two_way_link = bool_validator_factory('two_way_link')

# type

parse_type = preinstanced_parser_factory('type', ConnectionType, ConnectionType.unknown)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', ConnectionType)

# verified

parse_verified = bool_parser_factory('verified', False)
put_verified_into = bool_optional_putter_factory('verified', False)
validate_verified = bool_validator_factory('verified')

# visibility

parse_visibility = preinstanced_parser_factory('visibility', ConnectionVisibility, ConnectionVisibility.user_only)
put_visibility_into = preinstanced_optional_putter_factory('visibility', ConnectionVisibility.user_only)
validate_visibility = preinstanced_validator_factory('visibility', ConnectionVisibility)



from .action import *
from .fields import *
from .helpers import *
from .preinstanced import *


__all__ = (
    *action.__all__,
    *fields.__all__,
    *helpers.__all__,
    *preinstanced.__all__,
)



__all__ = ('create_partial_integration_from_id', )

from scarletio import export

from ...core import INTEGRATIONS

from .integration import Integration


@export
def create_partial_integration_from_id(integration_id, role_id = 0):
    """
    Creates an integration with the given id.
    
    If the integration already exists, returns that instead.
    
    Parameters
    ----------
    integration_id : `int`
        The unique identifier number of the integration.
    role_id : `int` = `0`, Optional
        The integration's role's identifier.
    
    Returns
    -------
    integration : ``Integration``
    """
    try:
        integration = INTEGRATIONS[integration_id]
    except KeyError:
        integration = Integration._create_with_role(integration_id, role_id)
        INTEGRATIONS[integration_id] = integration
    
    return integration



__all__ = ()

from .....env import API_VERSION

from ....emoji import create_unicode_emoji

from ..audit_log_change import AuditLogChange

from .shared import convert_color, convert_deprecated, convert_icon, convert_nothing, convert_permission


def convert_bool__separated(name, data):
    return convert_nothing('separated', data)


def convert_unicode_emoji(name, data):
    before = data.get('old_value', None)
    if (before is None):
        before = None
    else:
        before = create_unicode_emoji(before)
    
    after = data.get('new_value', None)
    if (after is None):
        after = None
    else:
        after = create_unicode_emoji(after)
    
    return AuditLogChange(name, before, after)


ROLE_CONVERTERS = {
    'color': convert_color,
    'hoist': convert_bool__separated,
    'icon_hash': convert_icon,
    'permissions': convert_deprecated if API_VERSION in (6, 7) else convert_permission,
    'permissions_new': convert_permission if API_VERSION in (6, 7) else convert_deprecated,
    'mentionable': convert_nothing,
    'name': convert_nothing,
    'position': convert_nothing,
    'unicode_emoji': convert_unicode_emoji,
}



__all__ = ('create_forum_tag_from_id', )

from ...core import FORUM_TAGS

from .fields import (
    put_emoji_into, put_moderated_into, put_name_into, validate_emoji, validate_moderated, validate_name
)
from .forum_tag import ForumTag


FORUM_TAG_FIELD_CONVERTERS = {
    'emoji': (validate_emoji, put_emoji_into),
    'moderated': (validate_moderated, put_moderated_into),
    'name': (validate_name, put_name_into),
}


def create_forum_tag_from_id(forum_tag_id):
    """
    Creates a forum tag from the given identifier.
    
    Parameters
    ----------
    forum_tag_id : `int`
        The forum tag's identifier.
    
    Returns
    -------
    forum_tag : ``ForumTag``
    """
    try:
        forum_tag = FORUM_TAGS[forum_tag_id]
    except KeyError:
        forum_tag = ForumTag._create_empty(forum_tag_id)
        FORUM_TAGS[forum_tag_id] = forum_tag
    
    return forum_tag



__all__ = ()

from ...field_parsers import flag_parser_factory, nullable_date_time_parser_factory
from ...field_putters import flag_optional_putter_factory, nullable_date_time_optional_putter_factory
from ...field_validators import flag_validator_factory, nullable_date_time_validator_factory

from .flags import ThreadProfileFlag

# flags

parse_flags = flag_parser_factory('flags', ThreadProfileFlag)
put_flags_into = flag_optional_putter_factory('flags', ThreadProfileFlag())
validate_flags = flag_validator_factory('flags', ThreadProfileFlag)

# joined_at

parse_joined_at = nullable_date_time_parser_factory('joined_at')
put_joined_at_into = nullable_date_time_optional_putter_factory('joined_at')
validate_joined_at = nullable_date_time_validator_factory('joined_at')



__all__ = ('InteractionMetadataMessageComponent',)

import warnings

from scarletio import copy_docs

from ...component import Component, ComponentType

from .base import InteractionMetadataBase
from .fields import (
    parse_component_type, parse_custom_id, parse_resolved, parse_values, put_component_type_into, put_custom_id_into,
    put_resolved_into, put_values_into, validate_component_type, validate_custom_id, validate_resolved, validate_values
)


class InteractionMetadataMessageComponent(InteractionMetadataBase):
    """
    Interaction metadata used when the interaction was triggered by an application command's auto completion.
    
    Parameters
    ----------
    component_type : ``ComponentType``
        The used component's type.
    
    custom_id : `None`, `str`
        Component or form interaction's custom identifier.
    
    resolved : `None`, ``Resolved``
        Contains the received entities.
    
    values : `None`, `tuple` of `str`
        Values selected by the user.
    """
    __slots__ = ('component_type', 'custom_id', 'resolved', 'values')
    
    @classmethod
    @copy_docs(InteractionMetadataBase._create_empty)
    def _create_empty(cls):
        self = object.__new__(cls)
        self.component_type = ComponentType.none
        self.custom_id = None
        self.resolved = None
        self.values = None
        return self
    
    
    @copy_docs(InteractionMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.component_type = self.component_type
        new.custom_id = self.custom_id
        
        resolved = self.resolved
        if (resolved is not None):
            resolved = resolved.copy()
        new.resolved = resolved
        
        values = self.values
        if (values is not None):
            values = (*values,)
        new.values = values
        
        return new
    
    
    @copy_docs(InteractionMetadataBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        # component_type
        try:
            component_type = keyword_parameters.pop('component_type')
        except KeyError:
            pass
        else:
            self.component_type = validate_component_type(component_type)
        
        # custom_id
        try:
            custom_id = keyword_parameters.pop('custom_id')
        except KeyError:
            pass
        else:
            self.custom_id = validate_custom_id(custom_id)
        
        # resolved
        try:
            resolved = keyword_parameters.pop('resolved')
        except KeyError:
            pass
        else:
            self.resolved = validate_resolved(resolved)
        
        # values
        try:
            values = keyword_parameters.pop('values')
        except KeyError:
            pass
        else:
            self.values = validate_values(values)
        
        
        InteractionMetadataBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase.from_data)
    def from_data(cls, data, interaction_event):
        self = object.__new__(cls)
        self.component_type = parse_component_type(data)
        self.custom_id = parse_custom_id(data)
        self.resolved = parse_resolved(data, interaction_event)
        self.values = parse_values(data)
        return self
    
    
    @copy_docs(InteractionMetadataBase.to_data)
    def to_data(self, *, defaults = False, interaction_event = None):
        data = {}
        put_component_type_into(self.component_type, data, defaults)
        put_custom_id_into(self.custom_id, data, defaults)
        put_resolved_into(self.resolved, data, defaults, interaction_event = interaction_event)
        put_values_into(self.values, data, defaults)
        return data
    
    
    @copy_docs(InteractionMetadataBase._put_attribute_representations_into)
    def _put_attribute_representations_into(self, repr_parts):
        # component_type
        repr_parts.append(' component_type = ')
        repr_parts.append(repr(self.component_type))
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id = ')
            repr_parts.append(repr(custom_id))
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            repr_parts.append(', resolved = ')
            repr_parts.append(repr(resolved))
        
        # values
        values = self.values
        if (values is not None):
            repr_parts.append(', values = ')
            repr_parts.append(repr(values))
        
        return True
    
    
    @copy_docs(InteractionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # component_type
        hash_value ^= self.component_type.value
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(self.custom_id)
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            hash_value ^= hash(resolved)
        
        # values
        values = self.values
        if (values is not None):
            hash_value ^= len(values)
            
            for option in values:
                hash_value ^= hash(option)
        
        return hash_value
    
    
    @copy_docs(InteractionMetadataBase.__eq__)
    def __eq__(self, other):
        other_type = type(other)
        if other_type is type(self):
            return self._is_equal_same_type(other)
        
        # Compare with components.
        if issubclass(other_type, Component):
            # Check `type` before `custom_id`
            
            # type
            if self.component_type is not other.type:
                return False
            
            # custom_id
            if self.custom_id != other.custom_id:
                return False
            
            return True
        
        return NotImplemented
    
    
    @copy_docs(InteractionMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # component_type
        if self.component_type is not other.component_type:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # resolved
        if self.resolved != other.resolved:
            return False
        
        # values
        if self.values != other.values:
            return False
        
        return True
    
    
    @property
    def options(self):
        """
        At the case of ``InteractionMetadataMessageComponent`` instances ``.options`` is deprecated and will be removed
        in 2023 February. Please use ``.values``.
        
        If you are accessing this field as a proxied attribute, you can ignore this warning.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.options` is deprecated and will be removed in 2023 February. '
                f'Please use `.values` instead. If you are accessing this field as a proxied attribute, '
                f'you can ignore this warning.'
            ),
            stacklevel = 2,
        )
        
        return self.values
    
    
    @property
    def type(self):
        """
        ``InteractionMetadataMessageComponent.type`` is deprecated and will be removed in 2023 February.
        Please use ``.component_type``.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.type` is deprecated and will be removed in 2023 February. '
                f'Please use `.component_type` instead.'
            ),
            stacklevel = 2,
        )
        
        return self.component_type



__all__ = ('InviteStage', )

from ..user import User


class InviteStage:
    """
    Represents an invite's stage.
    
    Attributes
    ----------
    participant_count : `int`
        The numbers of participants of the stage.
    participants : `tuple` of ``ClientUserBase``
        The users inside of the stage.
    speaker_count : int`
        The number of speakers in the stage.
    topic : `None`, `str`
        The stage's topic if any.
    """
    __slots__ = ('participant_count', 'participants', 'speaker_count', 'topic',)
    
    
    def __new__(cls, data, guild_id):
        """
        Creates a new ``InviteStage`` from the given data.
        
        Parameters
        ----------
        data : `str`
            Data received from Discord.
        guild_id : `int`
            The respective guild's identifier.
        """
        guild_profile_datas = data['members']
        users = tuple(
            User.from_data(guild_profile_data['user'], guild_profile_datas, guild_id)
            for guild_profile_data in guild_profile_datas
        )
        
        topic = data['topic']
        if (topic is not None) and (not topic):
            topic = None
        
        self = object.__new__(cls)
        self.participant_count = data['participant_count']
        self.speaker_count = data['speaker_count']
        self.participants = users
        self.topic = topic
        
        return self
    
    
    def __repr__(self):
        """Returns the invite stage's representation."""
        return f'<{self.__class__.__name__}>'



__all__ = ()

from .shared import convert_nothing, convert_snowflake


INVITE_CONVERTERS = {
    'channel_id': convert_snowflake,
    'code': convert_nothing,
    'inviter_id': convert_snowflake,
    'max_age': convert_nothing,
    'max_uses': convert_nothing,
    'temporary': convert_nothing,
    'uses': convert_nothing,
}
